from .trackmate_spot import TrackMateSpot


class TrackMateFrameSpots:
    """
    Parse TrackMate frame spots from xml file.
    """

    def __init__(self, trackmate_frame_spot, raw_video_shape: list[int]):
        self.spots: list[TrackMateSpot] = []
        for spot in trackmate_frame_spot["Spot"]:
            self.spots.append(TrackMateSpot(spot, raw_video_shape))

        self.frame = int(trackmate_frame_spot["@frame"])



import os
import sys
import torch
import imagej
import scyjava as sj
from cellpose import models

from ..utils.cell_track import CellTrack


class SegmentationTrackingFactory:
    """
    Class to perform cell segmentation and tracking.

    Args:
        model_path (str): path to the cellpose model
        augment (bool)
        cellprob_threshold (float)
        flow_threshold (float)
        gap_closing_max_distance_ratio (float): ratio of average spot size
        linking_max_distance_ratio  (float): ratio of average spot size
        max_frame_gap (int)
    """

    def __init__(
        self,
        model_path: str,
        augment=True,
        cellprob_threshold=0.0,
        flow_threshold=0.0,
        gap_closing_max_distance_ratio=0.5,
        linking_max_distance_ratio=1,
        max_frame_gap=CellTrack.max_frame_gap,
    ) -> None:
        self.model_path = model_path
        self.augment = augment
        self.cellprob_threshold = cellprob_threshold
        self.flow_threshold = flow_threshold
        self.gap_closing_max_distance_ratio = gap_closing_max_distance_ratio
        self.linking_max_distance_ratio = linking_max_distance_ratio
        self.max_frame_gap = max_frame_gap

    def perform_trackmate_tracking(
        self,
        video_path: str,
        fiji_path: str,
        save_folder: str,
        fast_mode: bool,
    ):
        """
        Use Trackmate to run both segmentation and tracking.
        """
        ij_instance = imagej.init(fiji_path, mode="interactive")

        bf_plugin = sj.jimport("loci.plugins.BF")
        importer_options = sj.jimport("loci.plugins.in.ImporterOptions")

        # Import TrackMate via scyjava
        track_mate_model = sj.jimport(
            "fiji.plugin.trackmate.Model"
        )  # class in charge of storing the data
        track_mate_settings = sj.jimport(
            "fiji.plugin.trackmate.Settings"
        )  # class storing the fields that will configure TrackMate and pilot how the data is created
        track_mate = sj.jimport("fiji.plugin.trackmate.TrackMate")
        logger = sj.jimport("fiji.plugin.trackmate.Logger")
        lap_utils = sj.jimport(
            "fiji.plugin.trackmate.tracking.jaqaman.LAPUtils"
        )
        sparse_lap_tracker_factory = sj.jimport(
            "fiji.plugin.trackmate.tracking.jaqaman.SparseLAPTrackerFactory"
        )
        file = sj.jimport("java.io.File")
        tm_xml_writer = sj.jimport("fiji.plugin.trackmate.io.TmXmlWriter")
        tracker_keys = sj.jimport("fiji.plugin.trackmate.tracking.TrackerKeys")
        track_mate_track_analyzer_provider = sj.jimport(
            "fiji.plugin.trackmate.providers.TrackAnalyzerProvider"
        )
        track_mate_edge_analyzer_provider = sj.jimport(
            "fiji.plugin.trackmate.providers.EdgeAnalyzerProvider"
        )

        if fast_mode:
            try:
                cellpose_detector_factory = sj.jimport(
                    "fiji.plugin.trackmate.cellpose.tbonte.CellposeDetectorFactory"
                )
                pretrained_model = sj.jimport(
                    "fiji.plugin.trackmate.cellpose.tbonte.CellposeSettings.PretrainedModel"
                )
            except TypeError:
                print(
                    "No Trackmate plugin found for fast mode. Using usual segmentation instead."
                )
                fast_mode = False

        if not fast_mode:
            cellpose_detector_factory = sj.jimport(
                "fiji.plugin.trackmate.cellpose.CellposeDetectorFactory"
            )
            pretrained_model = sj.jimport(
                "fiji.plugin.trackmate.cellpose.CellposeSettings.PretrainedModel"
            )

        # Skip if file already exists
        video_file_name = os.path.basename(video_path).split(".")[0]
        out_file = os.path.join(save_folder, f"{video_file_name}_model.xml")
        if os.path.exists(out_file):
            print(f"File {out_file} already exists. Skipping.")
            return

        # Get currently selected image
        options = importer_options()
        options.setColorMode(importer_options.COLOR_MODE_GRAYSCALE)
        options.setId(video_path)
        imps = bf_plugin.openImagePlus(options)
        imp = imps[0]

        # Swap Z and T dimensions if necessary
        dims = imp.getDimensions()
        if imp.dims[-1] == "Z":
            imp.setDimensions(
                sj.to_java(dims[2]), sj.to_java(dims[4]), sj.to_java(dims[3])
            )
            print("Swapping Z and T dimensions")

        # Get the average spot size to define max linking distances
        average_spot_size = models.CellposeModel(
            pretrained_model=[self.model_path]
        ).diam_labels

        # Create model object
        model = track_mate_model()

        # Send all messages to ImageJ log window.
        model.setLogger(logger.IJ_LOGGER)

        # Prepare settings object
        settings = track_mate_settings(imp)

        # Configure detector - We use the Strings for the keys
        settings.detectorFactory = cellpose_detector_factory()
        settings.detectorSettings["TARGET_CHANNEL"] = sj.to_java(3)
        settings.detectorSettings["OPTIONAL_CHANNEL_2"] = sj.to_java(0)
        settings.detectorSettings["CELLPOSE_PYTHON_FILEPATH"] = sys.executable
        settings.detectorSettings["CELLPOSE_MODEL_FILEPATH"] = self.model_path
        settings.detectorSettings["CELLPOSE_MODEL"] = pretrained_model.CUSTOM
        settings.detectorSettings["CELL_DIAMETER"] = sj.to_java(
            0.0, type="double"
        )
        settings.detectorSettings["USE_GPU"] = (
            True if torch.cuda.is_available() else False
        )
        settings.detectorSettings["SIMPLIFY_CONTOURS"] = True

        if fast_mode:
            settings.detectorSettings["FLOW_THRESHOLD"] = self.flow_threshold
            settings.detectorSettings["CELLPROB_THRESHOLD"] = (
                self.cellprob_threshold
            )
            settings.detectorSettings["AUGMENT"] = self.augment

        # Configure tracker
        settings.trackerFactory = sparse_lap_tracker_factory()
        settings.trackerSettings = (
            lap_utils.getDefaultSegmentSettingsMap()
        )  # almost good enough
        settings.trackerSettings["LINKING_MAX_DISTANCE"] = (
            self.linking_max_distance_ratio * average_spot_size
        )
        settings.trackerSettings["ALLOW_GAP_CLOSING"] = True
        settings.trackerSettings["GAP_CLOSING_MAX_DISTANCE"] = (
            self.gap_closing_max_distance_ratio * average_spot_size
        )
        settings.trackerSettings["MAX_FRAME_GAP"] = sj.to_java(
            self.max_frame_gap
        )
        settings.trackerSettings["ALLOW_TRACK_MERGING"] = False
        settings.trackerSettings["ALLOW_TRACK_SPLITTING"] = False
        settings.trackerSettings["LINKING_FEATURE_PENALTIES"] = (
            tracker_keys.DEFAULT_LINKING_FEATURE_PENALTIES
        )

        settings.initialSpotFilterValue = -1.0

        # Add useful track analyzers
        track_analyzer_provider = track_mate_track_analyzer_provider()
        settings.addTrackAnalyzer(
            track_analyzer_provider.getFactory("Track duration")
        )
        settings.addTrackAnalyzer(
            track_analyzer_provider.getFactory("Track index")
        )

        edge_analyzer_provider = track_mate_edge_analyzer_provider()
        settings.addEdgeAnalyzer(
            edge_analyzer_provider.getFactory("Edge target")
        )

        # Instantiate plugin
        trackmate = track_mate(model, settings)

        # Process
        process_ok = trackmate.checkInput()
        if not process_ok:
            sys.exit(str(trackmate.getErrorMessage()))

        process_ok = trackmate.process()
        if not process_ok:
            sys.exit(str(trackmate.getErrorMessage()))

        # Echo results with the logger we set at start:
        model.getLogger().log(str(model))

        out_file_model = file(save_folder, f"{video_file_name}_model.xml")
        writer = tm_xml_writer(out_file_model)
        writer.appendModel(model)
        writer.appendSettings(settings)
        writer.writeToFile()

        # Force exit
        ij_instance.dispose()



from __future__ import annotations
from abc import abstractmethod
from typing import TypeVar, Generic

from .cell_spot import CellSpot
from .mid_body_spot import MidBodySpot

T = TypeVar("T", MidBodySpot, CellSpot)


class Track(Generic[T]):
    """
    Class used for both cell and mid-body tracks.
    """

    def __init__(self, track_id: int) -> None:
        self.track_id = track_id

        self.spots: dict[int, T] = {}
        self.length = 0  # = len(self.spots)
        self.number_spots = (
            0  # can be different from length if we have a gap in the track
        )

    @staticmethod
    @abstractmethod
    def generate_tracks_from_spots(
        spots: dict[int, list[T]],
        linking_max_distance: int,
        gap_closing_max_distance: int,
    ) -> list[Track[T]]:
        """
        Generate tracks from spots.
        """



NAMES_DICTIONARY = {
    "No membrane cut, double MT": 0,
    "No membrane cut, simple MT": 1,
    "No membrane cut, no MT": 2,
    "Membrane cut, simple MT": 3,
    "Membrane cut, no MT": 4,
    "No membrane cut, double MT ?": 5,
    "No membrane cut, simple MT ?": 6,
    "No membrane cut, no MT ?": 7,
    "Membrane cut, simple MT ?": 8,
    "Membrane cut, no MT ?": 9,
}


def get_class_ids_after_first_mt_cut() -> list[int]:
    """
    Get ids of classes after first MT cut.
    """
    return [
        NAMES_DICTIONARY["No membrane cut, simple MT"],
        NAMES_DICTIONARY["No membrane cut, no MT"],
        NAMES_DICTIONARY["Membrane cut, simple MT"],
        NAMES_DICTIONARY["Membrane cut, no MT"],
        NAMES_DICTIONARY["No membrane cut, simple MT ?"],
        NAMES_DICTIONARY["No membrane cut, no MT ?"],
        NAMES_DICTIONARY["Membrane cut, simple MT ?"],
        NAMES_DICTIONARY["Membrane cut, no MT ?"],
    ]


def get_class_ids_after_second_mt_cut() -> list[int]:
    """
    Get ids of classes after second MT cut.
    """
    return [
        NAMES_DICTIONARY["No membrane cut, no MT"],
        NAMES_DICTIONARY["Membrane cut, no MT"],
        NAMES_DICTIONARY["No membrane cut, no MT ?"],
        NAMES_DICTIONARY["Membrane cut, no MT ?"],
    ]


def get_class_ids_after_first_membrane_cut() -> list[int]:
    """
    Get ids of classes after first membrane cut.
    """
    return [
        NAMES_DICTIONARY["Membrane cut, simple MT"],
        NAMES_DICTIONARY["Membrane cut, no MT"],
        NAMES_DICTIONARY["Membrane cut, simple MT ?"],
        NAMES_DICTIONARY["Membrane cut, no MT ?"],
    ]



import os
from typing import Optional
import pickle

from cut_detector.utils.mitosis_track import MitosisTrack
from cut_detector.data.tools import get_data_path


def main(mitoses_folder: Optional[str] = get_data_path("mitoses")):
    """
    Evaluate mid-body detection from annotated files.
    """
    # Load mitosis tracks
    mitoses_tracks: list[MitosisTrack] = []
    mitoses_files = os.listdir(mitoses_folder)
    for mitosis_file in mitoses_files:
        with open(os.path.join(mitoses_folder, mitosis_file), "rb") as f:
            mitosis_track: MitosisTrack = pickle.load(f)
            mitosis_track.adapt_deprecated_attributes()
            mitoses_tracks.append(mitosis_track)

    # Iterate over mitoses
    mb_detected, mb_not_detected = 0, 0
    wrong_detections = []
    for mitosis_track, mitosis_file in zip(mitoses_tracks, mitoses_files):
        # Ignore if triple division or not annotated mitosis
        if (
            len(mitosis_track.daughter_track_ids) != 1
            or mitosis_track.gt_mid_body_spots is None
        ):
            continue
        # Perform evaluation for current mitosis
        (
            is_correctly_detected,
            percent_detected,
            average_position_difference,
        ) = mitosis_track.evaluate_mid_body_detection()
        if is_correctly_detected:
            mb_detected += 1
        else:
            mb_not_detected += 1
            wrong_detections.append(
                {
                    "path": mitosis_file,
                    "percent_detected": percent_detected,
                    "average_position_difference": average_position_difference,
                }
            )

    # Print wrong detections
    if len(wrong_detections) > 0:
        print("\nWrong detections:")
        for wrong_detection in wrong_detections:
            print(
                f"{wrong_detection['path']}: detected {wrong_detection['percent_detected']}% with avg distance {wrong_detection['average_position_difference']}"
            )

    if (mb_detected + mb_not_detected) == 0:
        print("No mid-body detection evaluation possible.")
    else:
        print(
            f"\nMid-body detection evaluation: {mb_detected / (mb_detected + mb_not_detected) * 100:.2f}% | {mb_detected}/{mb_detected + mb_not_detected}"
        )


if __name__ == "__main__":
    main()



from __future__ import annotations
from typing import Callable, Optional
import numpy as np

from .track import Track
from .mid_body_spot import MidBodySpot


class MidBodyTrack(Track[MidBodySpot]):
    """
    Mid-body candidate track
    """

    max_frame_gap = 2

    def add_spot(self, spot: MidBodySpot) -> None:
        """
        Add spot to track.
        """
        self.spots[spot.frame] = spot
        spot.track_id = self.track_id
        self.length += 1
        # Add children recursively
        if spot.child_spot is not None:
            self.add_spot(spot.child_spot)

    def get_expected_distance(
        self, expected_positions: dict[int, list[int]], max_distance: float
    ) -> float:
        """
        Compute the average distance between mid-body expected positions and current
        track positions.
        """
        distances = []
        for frame, position in expected_positions.items():
            if frame not in self.spots:
                continue
            spot = self.spots[frame]
            distances.append(
                np.linalg.norm(spot.get_position() - np.array(position))
            )
        # If there are no frames in common, for sure track is not the right one
        if len(distances) == 0:
            return np.inf
        mean_distance = np.mean(distances)
        if mean_distance > max_distance:
            return np.inf
        return mean_distance

    @staticmethod
    def generate_tracks_from_spots(
        spots: dict[int, list[MidBodySpot]],
        linking_max_distance: int,
        gap_closing_max_distance: int = None,
        track_dist_metric: str | Callable = "sqeuclidean",
    ) -> list[MidBodyTrack]:
        """
        Generate tracks from spots.
        """
        max_frame_gap = MidBodyTrack.max_frame_gap
        raise NotImplementedError



from cnn_framework.utils.model_params.base_model_params import BaseModelParams
from cnn_framework.utils.dimensions import Dimensions


class BridgesMtCnnModelParams(BaseModelParams):
    """
    Bridges CNN model params.
    """

    def __init__(self):
        super().__init__("bridges_mt_cnn")

        self.input_dimensions = Dimensions(height=100, width=100)

        self.nb_classes = 3
        self.class_names = ["2 MT", "1 MT", "0 MT"]

        self.c_indexes = [0]
        self.z_indexes = [0]

        self.batch_size = 16

        self.encoder_name = "resnet18"



import os
import pickle
from typing import Optional
import numpy as np
from aicsimageio.writers import OmeTiffWriter

from ..factories.mid_body_detection_factory import MidBodyDetectionFactory

from ..utils.mitosis_track import MitosisTrack
from ..utils.trackmate_track import TrackMateTrack


def perform_mid_body_detection(
    raw_video: np.ndarray,
    video_name: str,
    exported_mitoses_dir: str,
    exported_tracks_dir: str,
    save_dir: Optional[str] = None,
    update_mitoses: bool = True,
):
    mitosis_tracks: list[MitosisTrack] = []
    # Iterate over "bin" files in exported_mitoses_dir
    for state_path in os.listdir(exported_mitoses_dir):
        # Ignore if not for current video
        if video_name not in state_path:
            continue
        # Load mitosis track
        with open(os.path.join(exported_mitoses_dir, state_path), "rb") as f:
            mitosis_track: MitosisTrack = pickle.load(f)
            mitosis_track.adapt_deprecated_attributes()

        # Add mitosis track to list
        mitosis_tracks.append(mitosis_track)

    # Load trackmate tracks
    trackmate_tracks: list[TrackMateTrack] = []
    # Iterate over "bin" files in exported_tracks_dir
    video_exported_tracks_dir = os.path.join(exported_tracks_dir, video_name)
    for state_path in os.listdir(video_exported_tracks_dir):
        # Load mitosis track
        with open(
            os.path.join(video_exported_tracks_dir, state_path), "rb"
        ) as f:
            trackmate_track: TrackMateTrack = pickle.load(f)
            trackmate_track.adapt_deprecated_attributes()
            trackmate_tracks.append(trackmate_track)

    # Generate movie for each mitosis and save
    mid_body_detector = MidBodyDetectionFactory()
    for i, mitosis_track in enumerate(mitosis_tracks):
        print(f"\nGenerate mitosis movies ({i+1}/{len(mitosis_tracks)})...")

        # Generate mitosis movie
        mitosis_movie, mask_movie = mitosis_track.generate_video_movie(
            raw_video
        )  # TYXC, TYX

        # Search for mid-body in mitosis movie
        mid_body_detector.update_mid_body_spots(
            mitosis_track, mitosis_movie, mask_movie, trackmate_tracks
        )

        # Save updated mitosis track
        if update_mitoses:
            daughter_track_ids = ",".join(
                [str(d) for d in mitosis_track.daughter_track_ids]
            )
            state_path = f"{video_name}_mitosis_{mitosis_track.id}_{mitosis_track.mother_track_id}_to_{daughter_track_ids}.bin"
            save_path = os.path.join(
                exported_mitoses_dir,
                state_path,
            )
            with open(save_path, "wb") as f:
                pickle.dump(mitosis_track, f)

        if save_dir:
            # Save mitosis movie
            final_mitosis_movie = mitosis_track.add_mid_body_movie(
                mitosis_movie, mask_movie
            )  # TYX C=C+1
            image_save_path = os.path.join(
                save_dir,
                f"{video_name}_mitosis_{mitosis_track.id}_{mitosis_track.mother_track_id}_to_{daughter_track_ids}.tiff",
            )
            # Transpose to match TCYX
            final_mitosis_movie = np.transpose(
                final_mitosis_movie, (0, 3, 1, 2)
            )
            OmeTiffWriter.save(
                final_mitosis_movie, image_save_path, dim_order="TCYX"
            )



import os
from typing import Optional
from cnn_framework.utils.readers.tiff_reader import TiffReader

from cut_detector.data.tools import get_data_path
from cut_detector.factories.mid_body_detection_factory import (
    MidBodyDetectionFactory,
)


def main(
    image_path: Optional[str] = get_data_path("mitosis_movies"),
    path_output: Optional[str] = get_data_path("mid_bodies"),
):
    # If image_path is a directory, take its first file
    if os.path.isdir(image_path):
        image_path = os.path.join(image_path, os.listdir(image_path)[0])

    # Read image
    image = TiffReader(image_path, respect_initial_type=True).image  # TCZYX

    mitosis_movie = image[:, :3, ...].squeeze()  # T C=3 YX
    mitosis_movie = mitosis_movie.transpose(0, 2, 3, 1)  # TYXC

    mask_movie = image[:, 3, ...].squeeze()  # TYX
    mask_movie = mask_movie.transpose(0, 1, 2)  # TYX

    # Search for mid-body in mitosis movie
    factory = MidBodyDetectionFactory()

    spots_candidates = factory.detect_mid_body_spots(
        mitosis_movie=mitosis_movie, mask_movie=mask_movie, mode="h_maxima"
    )  # mode = "bigfish" or "h_maxima" (default)
    factory.generate_tracks_from_spots(
        spots_candidates,
    )
    factory.save_mid_body_tracking(
        spots_candidates, mitosis_movie, path_output
    )


if __name__ == "__main__":
    main()



import os
from skimage import io

from cut_detector._widget import mitosis_track_generation
from cut_detector.data.tools import get_data_path
from cut_detector.widget_functions.mitosis_track_generation import (
    perform_mitosis_track_generation,
)


def test_open_track_generation_widget():
    # Just try to open the widget
    mitosis_track_generation()


def test_track_generation():

    # Add video
    video = video = io.imread(
        os.path.join(get_data_path("videos"), "example_video.tif")
    )  # TYXC

    # Open the widget
    mitosis_tracks = perform_mitosis_track_generation(
        video,
        "example_video",
        get_data_path("models"),
    )

    assert len(mitosis_tracks) == 1

    mitosis_track = mitosis_tracks[0]

    assert mitosis_track.daughter_track_ids == [0]
    assert mitosis_track.mother_track_id == 4
    assert 7 <= mitosis_track.key_events_frame["metaphase"] <= 9  # should be 8
    assert (
        10 <= mitosis_track.key_events_frame["cytokinesis"] <= 12
    )  # should be 11



import os
import numpy as np

from ..utils.bridges_classification.bridges_mt_model_manager import (
    BridgesMtModelManager,
)
from ..utils.bridges_classification.bridges_mt_cnn_model_params import (
    BridgesMtCnnModelParams,
)
from ..utils.tools import apply_hmm, perform_cnn_inference
from ..utils.bridges_classification.impossible_detection import (
    ImpossibleDetection,
)
from ..utils.bridges_classification.template_type import TemplateType
from ..utils.mitosis_track import MitosisTrack


class MtCutDetectionFactory:
    """
    Class to detect first and second MT cut.

    Args:
        margin (int): Number of pixels on each side of the midbody.

        # Bridges classification parameters

        coeff_height_peak (float): Coeff for the height of the peak.
        circle_radius (int): Middle circle radius.
        circle_min_ratio (float): Min ratio distance on circle between two peaks.
        coeff_width_peak (float): Coeff for the width of the peak.
        diff_radius (float): Distance between the min/max and middle circle.
        window_size (float): Window size for the peak detection (% of the circle perimeter).
        min_peaks_by_group (int): Minimum number of peaks by window for the peak detection.
        overlap (int): Overlap between two windows for the peak detection.
        template_type (TemplateType): Template for SVC model.

        # Light spot detection parameters

        intensity_threshold_light_spot (int): Intensity threshold for the light spot detection.
        h_maxima_light_spot (int): h for the h_maxima function (light spot detection).
        center_tolerance_light_spot (int): Center tolerance to not count the light spots that are
        too close to the center.
        min_percentage_light_spot (float): Minimum percentage of frames with light spots to
        consider the mitosis as a light spot mitosis.
        crop_size_light_spot (int): Size of the crop for the light spot detection.
        length_light_spot (int): Length of the video to check around the mt cut for light spot
        detection.
    """

    def __init__(
        self,
        margin=50,
        coeff_height_peak=1.071,
        circle_radius=11,
        circle_min_ratio=0.2025,
        coeff_width_peak=0.25358974,
        diff_radius=4.54,
        window_size=0.04836,
        min_peaks_by_group=4,
        overlap=4,
        intensity_threshold_light_spot=350,
        h_maxima_light_spot=130,
        center_tolerance_light_spot=5,
        min_percentage_light_spot=0.1,
        crop_size_light_spot=20,
        length_light_spot=3,
        template_type=TemplateType.ALL,
    ) -> None:
        self.margin = margin
        self.coeff_height_peak = coeff_height_peak
        self.circle_radius = circle_radius
        self.circle_min_ratio = circle_min_ratio
        self.coeff_width_peak = coeff_width_peak
        self.diff_radius = diff_radius
        self.window_size = window_size
        self.min_peaks_by_group = min_peaks_by_group
        self.overlap = overlap
        self.intensity_threshold_light_spot = intensity_threshold_light_spot
        self.h_maxima_light_spot = h_maxima_light_spot
        self.center_tolerance_light_spot = center_tolerance_light_spot
        self.min_percentage_light_spot = min_percentage_light_spot
        self.crop_size_light_spot = crop_size_light_spot
        self.length_light_spot = length_light_spot
        self.template_type = template_type

    @staticmethod
    def _is_bridges_classification_impossible(
        mitosis_track: MitosisTrack,
    ) -> bool:
        """
        Bridges classification is impossible if:
        - no midbody spots
        - more than 2 daughter tracks
        - nucleus is near border
        - no midbody spot after cytokinesis
        """

        if mitosis_track.is_near_border:
            mitosis_track.key_events_frame["first_mt_cut"] = (
                ImpossibleDetection.NEAR_BORDER
            )
            mitosis_track.key_events_frame["second_mt_cut"] = (
                ImpossibleDetection.NEAR_BORDER
            )
            return True

        if len(mitosis_track.daughter_track_ids) >= 2:
            mitosis_track.key_events_frame["first_mt_cut"] = (
                ImpossibleDetection.MORE_THAN_TWO_DAUGHTER_TRACKS
            )
            mitosis_track.key_events_frame["second_mt_cut"] = (
                ImpossibleDetection.MORE_THAN_TWO_DAUGHTER_TRACKS
            )
            return True

        if not mitosis_track.mid_body_spots:
            mitosis_track.key_events_frame["first_mt_cut"] = (
                ImpossibleDetection.NO_MID_BODY_DETECTED
            )
            mitosis_track.key_events_frame["second_mt_cut"] = (
                ImpossibleDetection.NO_MID_BODY_DETECTED
            )
            return True

        if (
            max(mitosis_track.mid_body_spots.keys())
            < mitosis_track.key_events_frame["cytokinesis"]
        ):
            mitosis_track.key_events_frame["first_mt_cut"] = (
                ImpossibleDetection.NO_MID_BODY_DETECTED_AFTER_CYTOKINESIS
            )
            mitosis_track.key_events_frame["second_mt_cut"] = (
                ImpossibleDetection.NO_MID_BODY_DETECTED_AFTER_CYTOKINESIS
            )
            return True

        return False

    def update_mt_cut_detection(
        self,
        mitosis_tracks: list[MitosisTrack],
        video: np.ndarray,
        hmm_bridges_parameters_file: str,
        bridges_mt_cnn_model_path: str,
        debug_mode=False,
    ):
        """
        Update micro-tubules cut detection using bridges classification.
        Expect TYXC video.
        """
        # Run CNN classification
        list_class_bridges = self.classify_bridges(
            mitosis_tracks, video, bridges_mt_cnn_model_path
        )

        # Debug use
        results = {
            "list_class_bridges": list_class_bridges,
            "list_class_bridges_after_hmm": {},
            "crops": {},
        }

        # Read HMM parameters
        if not os.path.exists(hmm_bridges_parameters_file):
            raise FileNotFoundError(
                f"File {hmm_bridges_parameters_file} not found"
            )
        hmm_parameters = np.load(hmm_bridges_parameters_file)

        # Check if classification is impossible and smooth
        for mitosis_track in mitosis_tracks:
            classification_impossible = (
                self._is_bridges_classification_impossible(mitosis_track)
            )

            if classification_impossible:
                continue

            # Perform classification...
            ordered_mb_frames = sorted(mitosis_track.mid_body_spots.keys())
            first_mb_frame = ordered_mb_frames[0]
            first_frame = max(
                first_mb_frame,
                mitosis_track.key_events_frame["cytokinesis"] - 2,
            )  # -2 because cytokinesis frame may be a bit too late

            if debug_mode:  # save crops only in debug
                results["crops"][mitosis_track.id] = (
                    mitosis_track.get_bridge_images(video, self.margin)
                )

            # Make sure cytokinesis bridge is detected as class 0: "no MT cut"
            relative_cytokinesis_frame = (
                mitosis_track.key_events_frame["cytokinesis"] - first_frame
            )
            if (
                relative_cytokinesis_frame < 0
                or results["list_class_bridges"][mitosis_track.id][
                    relative_cytokinesis_frame
                ]
                != 0
            ):
                mitosis_track.key_events_frame["first_mt_cut"] = (
                    ImpossibleDetection.MT_CUT_AT_CYTOKINESIS
                )
                mitosis_track.key_events_frame["second_mt_cut"] = (
                    ImpossibleDetection.MT_CUT_AT_CYTOKINESIS
                )
                continue

            # Correct the sequence with HMM
            results["list_class_bridges_after_hmm"][mitosis_track.id] = (
                apply_hmm(
                    hmm_parameters,
                    results["list_class_bridges"][mitosis_track.id],
                )
            )

            # Get index of first element > 0 in sequence (first MT cut)
            first_mt_cut_frame_rel = next(
                (
                    i
                    for i, x in enumerate(
                        results["list_class_bridges_after_hmm"][
                            mitosis_track.id
                        ]
                    )
                    if x > 0
                ),
                -1,
            )

            # Ignore if no MT cut detected
            if first_mt_cut_frame_rel == -1:
                mitosis_track.key_events_frame["first_mt_cut"] = (
                    ImpossibleDetection.NO_CUT_DETECTED
                )
                mitosis_track.key_events_frame["second_mt_cut"] = (
                    ImpossibleDetection.NO_CUT_DETECTED
                )
                continue

            first_mt_cut_frame_abs = first_frame + first_mt_cut_frame_rel
            if mitosis_track.light_spot_detected(
                video,
                first_mt_cut_frame_abs,
                self.length_light_spot,
                self.crop_size_light_spot,
                self.h_maxima_light_spot,
                self.intensity_threshold_light_spot,
                self.center_tolerance_light_spot,
                self.min_percentage_light_spot,
            ):
                mitosis_track.key_events_frame["first_mt_cut"] = (
                    ImpossibleDetection.LIGHT_SPOT
                )
                mitosis_track.key_events_frame["second_mt_cut"] = (
                    ImpossibleDetection.LIGHT_SPOT
                )
                continue

            # Update mitosis track accordingly
            mitosis_track.key_events_frame["first_mt_cut"] = (
                first_mt_cut_frame_abs
            )

            # Get index of first element > 1 in sequence (second MT cut)
            second_mt_cut_frame_rel = next(
                (
                    i
                    for i, x in enumerate(
                        results["list_class_bridges_after_hmm"][
                            mitosis_track.id
                        ]
                    )
                    if x > 1
                ),
                -1,
            )

            # get the frame of the second MT cut
            if second_mt_cut_frame_rel == -1:
                mitosis_track.key_events_frame["second_mt_cut"] = (
                    ImpossibleDetection.NO_CUT_DETECTED
                )
                continue

            second_mt_cut_frame_abs = first_frame + second_mt_cut_frame_rel

            # Update mitosis track accordingly
            mitosis_track.key_events_frame["second_mt_cut"] = (
                second_mt_cut_frame_abs
            )

            if debug_mode:
                # In debug, expect only one mitosis track
                return results

        return results

    def classify_bridges(
        self,
        mitosis_tracks: list[MitosisTrack],
        video: np.ndarray,
        bridges_mt_cnn_model_path: str,
    ) -> dict[int, list[int]]:
        """
        Classify bridges using a CNN model.
        """
        # Get bridge crops
        crops = {}
        for mitosis_track in mitosis_tracks:
            raw_crops = mitosis_track.get_bridge_images(video, self.margin)
            first_channel_crops = [
                np.expand_dims(raw_crop[0], axis=0) for raw_crop in raw_crops
            ]
            crops[mitosis_track.id] = first_channel_crops

        # Perform classification
        crops_list = [
            crop for crop_list in crops.values() for crop in crop_list
        ]
        predictions = perform_cnn_inference(
            model_path=bridges_mt_cnn_model_path,
            images=crops_list,
            cnn_model_params=BridgesMtCnnModelParams,
            model_manager=BridgesMtModelManager,
        )

        # Create prediction dictionary
        mitosis_predictions = {}
        for mitosis_track_id, crop_list in crops.items():
            mitosis_predictions[mitosis_track_id] = predictions[
                : len(crop_list)
            ]
            predictions = predictions[len(crop_list) :]

        return mitosis_predictions



class BoxDimensions:
    """
    A class to store the dimensions of a box.
    """

    def __init__(self, min_x=None, max_x=None, min_y=None, max_y=None):
        self.min_x = min_x
        self.max_x = max_x
        self.min_y = min_y
        self.max_y = max_y

    def is_empty(self):
        return self.min_x is None

    def update(self, min_x, max_x, min_y, max_y):
        if self.min_x is None or min_x < self.min_x:
            self.min_x = min_x
        if self.max_x is None or max_x > self.max_x:
            self.max_x = max_x
        if self.min_y is None or min_y < self.min_y:
            self.min_y = min_y
        if self.max_y is None or max_y > self.max_y:
            self.max_y = max_y
        # Protect against wrong box dimensions
        assert self.min_x <= self.max_x
        assert self.min_y <= self.max_y

    def update_from_box_dimensions(self, box_dimensions):
        self.update(
            box_dimensions.min_x, box_dimensions.max_x, box_dimensions.min_y, box_dimensions.max_y
        )

    def overlaps(self, box_dimensions):
        return (
            self.min_x <= box_dimensions.max_x
            and self.max_x >= box_dimensions.min_x
            and self.min_y <= box_dimensions.max_y
            and self.max_y >= box_dimensions.min_y
        )



# Metaphase links
FRAMES_AROUND_METAPHASE = (
    10  # look for metaphase candidate spots in this range
)

INTERPHASE_INDEX = 0  # interphase index for CNN classification model
METAPHASE_INDEX = 1  # metaphase index for CNN classification model

# Mid-body tracking
CYTOKINESIS_DURATION = (
    20  # number of frames to look for mid-body in between cells
)

# Minimum distance to border to consider is_near_border = true
MINIMUM_DISTANCE_TO_BORDER = 20



import os
import urllib.request

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


def get_model_path(model_name: str) -> None:
    """
    Returns absolute path to model file.
    """
    if model_name == "segmentation":
        files = ["segmentation_model"]
    elif model_name == "metaphase_cnn":
        model_name = "metaphase_cnn_v009"
        files = ["mean_std.json", "metaphase_cnn.pt", "parameters.csv"]
    elif model_name == "hmm":
        files = ["hmm_metaphase_parameters.npz", "hmm_bridges_parameters.npz"]
    elif model_name == "svc_bridges":
        files = ["scaler.pkl", "model.pkl"]
    elif model_name == "bridges_mt_cnn":
        model_name = "bridges_mt_cnn_v009"
        files = ["mean_std.json", "bridges_mt_cnn.pt", "parameters.csv"]
    else:
        raise ValueError(f"Unknown model name: {model_name}")

    local_path = os.path.join(CURRENT_DIR, model_name)
    os.makedirs(local_path, exist_ok=True)

    for file in files:
        file_local_path = os.path.join(local_path, file)
        if not os.path.exists(file_local_path):
            print(f"Downloading data {model_name}...")
            urllib.request.urlretrieve(
                f"https://raw.githubusercontent.com//15bonte/cut-detector-models/main/models/{model_name}/{file}",
                file_local_path,
            )

    return local_path



import os
from pathlib import Path
from typing import Optional

from magicgui import magic_factory
import tempfile


from .utils.tools import re_organize_channels

from .widget_functions.tracking import perform_tracking
from .widget_functions.mid_body_detection import perform_mid_body_detection
from .widget_functions.mitosis_track_generation import (
    perform_mitosis_track_generation,
)
from .widget_functions.mt_cut_detection import perform_mt_cut_detection
from .widget_functions.save_results import perform_results_saving


@magic_factory(
    call_button="Run Whole Process",
    layout="vertical",
    fiji_dir=dict(
        widget_type="FileEdit",
        label="Fiji path (likely named Fiji.app): ",
        mode="d",
    ),
    default_model_check_box=dict(
        widget_type="CheckBox",
        text="Use default segmentation model?",
        value=True,
    ),
    segmentation_model=dict(
        widget_type="FileEdit",
        label="If not checked, cellpose segmentation model: ",
    ),
    fast_mode_check_box=dict(
        widget_type="CheckBox", text="Enable fast mode?", value=False
    ),
    save_check_box=dict(
        widget_type="CheckBox", text="Save cell divisions movies?", value=False
    ),
    movies_save_dir=dict(
        widget_type="FileEdit",
        label="If checked, directory to save division movies: ",
        mode="d",
    ),
    results_save_dir=dict(
        widget_type="FileEdit",
        label="Directory to save results: ",
        mode="d",
    ),
)
def whole_process(
    img_layer: "napari.layers.Image",
    img_viewer: "napari.Viewer",
    fiji_dir: str,
    default_model_check_box: bool,
    segmentation_model: str,
    fast_mode_check_box: bool,
    save_check_box: bool,
    movies_save_dir: str,
    results_save_dir: str,
):
    # Make sure we are working with the same image (same name)
    video_path = None
    for layer in img_viewer.layers:
        local_video_path = layer.source.path
        if img_layer.name == os.path.basename(local_video_path).split(".")[0]:
            # This is the same image, we keep the link
            video_path = local_video_path
            break
    assert video_path is not None

    # Create temporary folders
    xml_model_dir_instance = tempfile.TemporaryDirectory()
    xml_model_dir = xml_model_dir_instance.name
    exported_mitoses_dir_instance = tempfile.TemporaryDirectory()
    exported_mitoses_dir = exported_mitoses_dir_instance.name
    exported_tracks_dir_instance = tempfile.TemporaryDirectory()
    exported_tracks_dir = exported_tracks_dir_instance.name

    # Segmentation and tracking
    video_path = str(Path(video_path))
    segmentation_model = str(Path(segmentation_model))
    perform_tracking(
        video_path,
        fiji_dir,
        xml_model_dir,
        segmentation_model if not default_model_check_box else None,
        fast_mode_check_box,
    )

    raw_video = re_organize_channels(img_layer.data)  # TXYC

    # Mitosis track_generation
    perform_mitosis_track_generation(
        raw_video,
        img_layer.name,
        xml_model_dir,
        exported_mitoses_dir,
        exported_tracks_dir,
    )

    # Mid-body detection
    perform_mid_body_detection(
        raw_video,
        img_layer.name,
        exported_mitoses_dir,
        exported_tracks_dir,
        movies_save_dir if save_check_box else None,
    )

    # MT cut detection
    perform_mt_cut_detection(raw_video, img_layer.name, exported_mitoses_dir)

    # Results saving
    perform_results_saving(exported_mitoses_dir, save_dir=results_save_dir)

    # Delete temporary folders
    xml_model_dir_instance.cleanup()
    exported_mitoses_dir_instance.cleanup()
    exported_tracks_dir_instance.cleanup()

    print("\nWhole process finished with success!")


@magic_factory(
    call_button="Run Segmentation and Tracking",
    layout="vertical",
    fiji_dir=dict(
        widget_type="FileEdit",
        label="Fiji path (likely named Fiji.app): ",
        mode="d",
    ),
    xml_model_dir=dict(
        widget_type="FileEdit",
        label=".xml models directory to save: ",
        mode="d",
    ),
    default_model_check_box=dict(
        widget_type="CheckBox",
        text="Use default segmentation model?",
        value=True,
    ),
    segmentation_model=dict(
        widget_type="FileEdit",
        label="If not checked, cellpose segmentation model: ",
    ),
    fast_mode_check_box=dict(
        widget_type="CheckBox", text="Enable fast mode?", value=False
    ),
)
def segmentation_tracking(
    img_viewer: "napari.Viewer",
    fiji_dir: str,
    xml_model_dir: str,
    default_model_check_box: bool,
    segmentation_model: str,
    fast_mode_check_box: bool,
):
    # Make sure there is only one image loaded as no layer has to be provided
    assert len(img_viewer.layers) == 1
    # Convert video path from windows to linux
    video_path = img_viewer.layers[0].source.path
    video_path = str(Path(video_path))
    segmentation_model = str(Path(segmentation_model))
    perform_tracking(
        video_path,
        fiji_dir,
        xml_model_dir,
        segmentation_model if not default_model_check_box else None,
        fast_mode_check_box,
    )
    print("\nSegmentation and tracking finished with success!")


@magic_factory(
    call_button="Run Mitosis Track Generation",
    layout="vertical",
    xml_model_dir=dict(
        widget_type="FileEdit",
        label=".xml models directory: ",
        mode="d",
    ),
    mitoses_save_dir=dict(
        widget_type="FileEdit",
        label="Directory to save .bin mitoses: ",
        mode="d",
    ),
    tracks_save_dir=dict(
        widget_type="FileEdit",
        label="Directory to save .bin tracks: ",
        mode="d",
    ),
)
def mitosis_track_generation(
    img_layer: "napari.layers.Image",
    xml_model_dir: str,
    mitoses_save_dir: Optional[str],
    tracks_save_dir: Optional[str],
):
    raw_video = re_organize_channels(img_layer.data)  # TXYC
    perform_mitosis_track_generation(
        raw_video,
        img_layer.name,
        xml_model_dir,
        mitoses_save_dir,
        tracks_save_dir,
    )
    print("\nMitosis tracks generated with success!")


@magic_factory(
    call_button="Run Mid-body Detection",
    layout="vertical",
    exported_mitoses_dir=dict(
        widget_type="FileEdit",
        label="Saved .bin mitoses directory: ",
        mode="d",
    ),
    exported_tracks_dir=dict(
        widget_type="FileEdit",
        label="Saved .bin tracks directory: ",
        mode="d",
    ),
    save_check_box=dict(
        widget_type="CheckBox", text="Save cell divisions movies?", value=False
    ),
    save_dir=dict(
        widget_type="FileEdit",
        label="If checked, directory to save division movies: ",
        mode="d",
    ),
)
def mid_body_detection(
    img_layer: "napari.layers.Image",
    exported_mitoses_dir: str,
    exported_tracks_dir: str,
    save_check_box: bool,
    save_dir: str,
):
    raw_video = re_organize_channels(img_layer.data)  # TXYC
    perform_mid_body_detection(
        raw_video,
        img_layer.name,
        exported_mitoses_dir,
        exported_tracks_dir,
        save_dir if save_check_box else None,
    )
    print("\nMid-body detection finished with success!")


@magic_factory(
    call_button="Run MT Cut Detection",
    layout="vertical",
    exported_mitoses_dir=dict(
        widget_type="FileEdit",
        label="Saved .bin mitoses directory: ",
        mode="d",
    ),
)
def micro_tubules_cut_detection(
    img_layer: "napari.layers.Image", exported_mitoses_dir: str
):
    raw_video = re_organize_channels(img_layer.data)  # TXYC
    perform_mt_cut_detection(
        raw_video,
        img_layer.name,
        exported_mitoses_dir,
    )
    print("\nMicro-tubules cut detection finished with success!")


@magic_factory(
    call_button="Save results",
    layout="vertical",
    exported_mitoses_dir=dict(
        widget_type="FileEdit",
        label="Saved .bin mitoses directory: ",
        mode="d",
    ),
    results_save_dir=dict(
        widget_type="FileEdit",
        label="Directory to save results: ",
        mode="d",
    ),
)
def results_saving(
    exported_mitoses_dir: str,
    results_save_dir: str,
):
    perform_results_saving(exported_mitoses_dir, save_dir=results_save_dir)
    print("\nResults saved with success!")



from .box_dimensions import BoxDimensions


class BoxDimensionsDln(BoxDimensions):
    """
    A class to store the dimensions of a box and the delaunay triangulation of the box.
    """

    def __init__(self):
        super().__init__(None, None, None, None)
        self.dln = None



import os
from typing import Optional, Union
import pickle
import numpy as np
from matplotlib import pyplot as plt

from cnn_framework.utils.display_tools import display_progress

from ..utils.cell_spot import CellSpot
from ..utils.trackmate_track import TrackMateTrack
from ..factories.tracks_merging_factory import TracksMergingFactory
from ..models.tools import get_model_path
from ..utils.mitosis_track import MitosisTrack


def plot_predictions_evolution(
    raw_spots: list[CellSpot],
    raw_tracks: list[TrackMateTrack],
    mitosis_tracks: list[MitosisTrack],
) -> None:
    # Spots detected by Cellpose
    detected_spots = {}
    for raw_spot in raw_spots:
        if raw_spot.frame not in detected_spots:
            detected_spots[raw_spot.frame] = 0
        detected_spots[raw_spot.frame] += 1

    # Mitoses identified by current method, minus ending tracks
    detected_mitoses = {frame: detected_spots[0] for frame in detected_spots}
    max_frame = max(detected_spots.keys())
    for mitosis_track in mitosis_tracks:
        for frame in range(mitosis_track.metaphase_frame, max_frame + 1):
            detected_mitoses[frame] += len(mitosis_track.daughter_track_ids)

    min_frame, max_frame = 0, max(detected_spots.keys())

    # Tracks identified by TrackMate
    detected_tracks = {}
    metaphase_spots = []
    for track in raw_tracks:
        metaphase_spots = metaphase_spots + [
            metaphase_spot.frame for metaphase_spot in track.metaphase_spots
        ]
        for frame in range(track.start, track.stop + 1):
            if frame not in detected_tracks:
                detected_tracks[frame] = 0
            detected_tracks[frame] += 1

    data_spots = [detected_spots[0]] + [
        detected_spots[i] if i in detected_spots else 0
        for i in range(min_frame, max_frame)
    ]
    data_tracks = [detected_tracks[0]] + [
        detected_tracks[i] if i in detected_tracks else 0
        for i in range(min_frame, max_frame)
    ]
    data_mitoses = [detected_mitoses[0]] + [
        detected_mitoses[i] if i in detected_mitoses else 0
        for i in range(min_frame, max_frame)
    ]

    _, ax = plt.subplots(1, 1, figsize=(15, 5))
    ax.step(
        list(range(min_frame, max_frame + 1)), data_spots, "r", linewidth=8.0
    )
    ax.step(
        list(range(min_frame, max_frame + 1)), data_tracks, "g", linewidth=8.0
    )
    ax.step(
        list(range(min_frame, max_frame + 1)), data_mitoses, "b", linewidth=8.0
    )

    ending_tracks = [track.stop + 1 for track in raw_tracks]

    # Plot first for legend
    ax.axvline(-10, color="y", linewidth=2.0)
    ax.axvline(-10, color="c", linewidth=2.0)
    ax.axvline(-10, color="k", linewidth=2.0, linestyle="--")

    # Potential mitoses
    metaphase_spots = list(set(metaphase_spots))
    for metaphase_spot_frame in metaphase_spots:
        ax.axvline(metaphase_spot_frame, color="y", linewidth=2.0)

    # Actual mitoses
    actual_mitoses = [mitosis.metaphase_frame for mitosis in mitosis_tracks]
    for actual_mitosis in actual_mitoses:
        ax.axvline(actual_mitosis, color="c", linewidth=2.0)

    # Ending tracks
    for ending_track_frame in ending_tracks:
        ax.axvline(
            ending_track_frame, color="k", linewidth=2.0, linestyle="--"
        )

    ax.set_xlim(min_frame, max_frame)

    ax.legend(
        [
            "Spots",
            "Tracks",
            "Mitoses evolution",
            "Potential mitoses",
            "Actual mitoses",
            "Ending tracks",
        ],
        loc="best",
    )
    plt.show()


def perform_mitosis_track_generation(
    raw_video: np.ndarray,
    video_name: str,
    xml_model_dir: str,
    mitoses_save_dir: Optional[str] = None,
    tracks_save_dir: Optional[str] = None,
    metaphase_model_path: Optional[str] = get_model_path("metaphase_cnn"),
    hmm_metaphase_parameters_file: Optional[str] = os.path.join(
        get_model_path("hmm"), "hmm_metaphase_parameters.npz"
    ),
    predictions_file: Optional[str] = None,
    only_predictions_update: bool = False,
    plot_evolution: bool = False,
) -> Union[list[MitosisTrack], None]:
    """
    Perform mitosis track generation.
    """
    # Create save_dir if not exists
    if mitoses_save_dir is not None and not os.path.exists(mitoses_save_dir):
        os.makedirs(mitoses_save_dir)

    # Create video tracks save dir if not exists
    if tracks_save_dir is not None:
        video_tracks_save_dir = os.path.join(tracks_save_dir, video_name)
        if not os.path.exists(video_tracks_save_dir):
            os.makedirs(video_tracks_save_dir)
    else:
        video_tracks_save_dir = None

    # Create factory instance, where useful functions are defined
    tracks_merging_factory = TracksMergingFactory()

    # Read useful information from xml file
    xml_model_path = os.path.join(xml_model_dir, f"{video_name}_model.xml")
    cell_tracks, cell_spots = tracks_merging_factory.read_trackmate_xml(
        xml_model_path, raw_video.shape
    )

    # Missing xml file case
    if cell_tracks is None:
        return None

    # Get dictionary of TrackMate spots (from xml file) for each track and detect metaphase spots
    tracks_merging_factory.pre_process_spots(
        cell_tracks,
        cell_spots,
        raw_video,
        metaphase_model_path,
        hmm_metaphase_parameters_file,
        predictions_file,
        video_name,
        only_predictions_update,
    )

    # If the goal is only to update predictions file, stop here
    if only_predictions_update:
        return None

    print("\nGet tracks to merge...")

    # Plug tracks occurring at frame>0 to closest metaphase
    mitosis_tracks = tracks_merging_factory.get_tracks_to_merge(cell_tracks)

    # Plot predictions evolution
    if plot_evolution:
        plot_predictions_evolution(cell_spots, cell_tracks, mitosis_tracks)

    # Update useful attributes for each track
    for i, mitosis_track in enumerate(mitosis_tracks):
        mitosis_track.id = i
        mitosis_track.update_mitosis_start_end(cell_tracks, mitosis_tracks)
        mitosis_track.update_key_events_frame(cell_tracks)
        mitosis_track.update_mitosis_position_dln(cell_tracks)
        mitosis_track.update_is_near_border(raw_video)

        # Save mitosis track
        if mitoses_save_dir is not None:
            daughter_track_ids = ",".join(
                [str(d) for d in mitosis_track.daughter_track_ids]
            )
            state_path = f"{video_name}_mitosis_{i}_{mitosis_track.mother_track_id}_to_{daughter_track_ids}.bin"
            save_path = os.path.join(
                mitoses_save_dir,
                state_path,
            )
            with open(save_path, "wb") as f:
                pickle.dump(mitosis_track, f)

        display_progress(
            "Mitosis tracks generation:",
            i + 1,
            len(mitosis_tracks),
            additional_message=f"Mitosis {i + 1}/{len(mitosis_tracks)}",
        )

    # Save updated cell tracks
    if video_tracks_save_dir is not None:
        for cell_track in cell_tracks:
            state_path = f"track_{cell_track.track_id}.bin"
            save_path = os.path.join(
                video_tracks_save_dir,
                state_path,
            )
            with open(save_path, "wb") as f:
                pickle.dump(cell_track, f)

    return mitosis_tracks



from __future__ import annotations

from typing import Optional, Tuple
from scipy.spatial import ConvexHull, Delaunay
import numpy as np


from ..constants.tracking import (
    FRAMES_AROUND_METAPHASE,
    INTERPHASE_INDEX,
    METAPHASE_INDEX,
)
from .track import Track
from .box_dimensions_dln import BoxDimensionsDln
from .box_dimensions import BoxDimensions
from .cell_spot import CellSpot


def get_whole_box_dimensions_dln(
    tracks: list[CellTrack], frame: int
) -> Tuple[BoxDimensionsDln, list[list[int]]]:
    """
    Merge different tracks.

    Parameters
    ----------
    predictions : list[int]
        list of predictions for each frame of the track.

    Returns
    -------
    BoxDimensionsDln : Box dimension of merged tracks.
    list[list[int]] : List of box dimension coordinates for all tracks.

    """
    box_dimensions_dln = BoxDimensionsDln()
    track_frame_points = []

    # For all tracks: mother and daughter(s)
    for track in tracks:
        if frame in track.spots:
            current_spot = track.spots[frame]
            track_frame_points = track_frame_points + current_spot.spot_points
            box_dimensions_dln.update(
                current_spot.abs_min_x,
                current_spot.abs_max_x,
                current_spot.abs_min_y,
                current_spot.abs_max_y,
            )
    return box_dimensions_dln, track_frame_points


class CellTrack(Track[CellSpot]):
    """
    Cell track.
    """

    max_frame_gap = 3

    def __init__(
        self, track_id: int, track_spots_ids: set[int], start: int, stop: int
    ) -> None:
        super().__init__(track_id)
        self.track_spots_ids = track_spots_ids
        self.start = start
        self.stop = stop

        self.metaphase_spots: list[CellSpot] = []

    def update_metaphase_spots(self, predictions: list[int]) -> None:
        """
        Populate metaphase_spots with candidates.

        Parameters
        ----------
        predictions : list[int]
            list of predictions for each frame of the track.

        Returns
        -------
        None.

        """
        for idx, frame in enumerate(sorted(self.spots.keys())):
            self.spots[frame].predicted_phase = predictions[idx]

        # Store last metaphase spot of each group
        metaphase_finished = False
        for frame in range(self.start, self.stop + 1):
            # Ignore first spots of cell as they are metaphase only if end of previous metaphase
            if predictions[frame - self.start] == INTERPHASE_INDEX:
                metaphase_finished = True
            if not metaphase_finished:
                continue

            # From this point, get metaphase spots
            if (
                frame in self.spots  # current frame contains a spot
                and predictions[frame - self.start]
                == METAPHASE_INDEX  # current spot is in metaphase
                and frame != self.stop  # current spot is not last spot
                and (
                    predictions[frame - self.start + 1] != METAPHASE_INDEX
                    and frame in self.spots
                )  # next frame is not a spot in metaphase
            ):
                self.metaphase_spots.append(self.spots[frame])

    def has_close_metaphase(self, spot: CellSpot, target_frame: int) -> bool:
        """
        Parameters
        ----------
        spot : CellSpot
            Only used to update corresponding metaphase spot.

        Returns
        -------
        bool : True if corresponding track contains one metaphase spot close to target frame.

        """
        # Look for metaphase spot
        for metaphase_spot in self.metaphase_spots:
            if (
                abs(metaphase_spot.frame - target_frame)
                < FRAMES_AROUND_METAPHASE
            ):
                # Mother track found!
                spot.corresponding_metaphase_spot = metaphase_spot
                return True
        return False

    def compute_metaphase_iou(self, daughter_track: CellTrack) -> float:
        """
        Get intersection between daughter area at first frame and self area at previous frame.
        Get self area at previous frame. Returns the quotient.

        Ideally, it should be close to 0.5 as a daughter cell should occupy 50% of the area
        of the mother cell.

        May be improved by checking overlap of areas instead of convex hulls.

        Parameters
        ----------
        daughter_track : CellTrack
            Potential daughter track.

        Returns
        -------
        float : Intersection Over Union.

        """

        daughter_track_first_frame = min(daughter_track.spots.keys())

        # If current track starts at first frame, ignore as it cannot be a mother track
        if self.start == daughter_track_first_frame:
            return -1

        # Compute two regions
        self_previous_region = self.compute_dln_from_tracks(
            daughter_track_first_frame - 1, relative=False
        )
        daughter_region = daughter_track.compute_dln_from_tracks(
            daughter_track_first_frame, relative=False
        )

        local_shape = (
            max(self_previous_region.max_y, daughter_region.max_y),
            max(self_previous_region.max_x, daughter_region.max_x),
        )

        indices = np.stack(
            np.indices(local_shape),
            axis=-1,
        )

        self_previous_out_idx = np.nonzero(
            self_previous_region.dln.find_simplex(indices) + 1
        )
        self_previous_region_mask = np.zeros(local_shape, dtype=bool)
        self_previous_region_mask[self_previous_out_idx] = True

        daughter_out_idx = np.nonzero(
            daughter_region.dln.find_simplex(indices) + 1
        )
        daughter_region_mask = np.zeros(local_shape, dtype=bool)
        daughter_region_mask[daughter_out_idx] = True

        # Compute intersection of both regions
        overlap = np.sum(self_previous_region_mask * daughter_region_mask)

        # Compute previous region area
        previous_area = np.sum(self_previous_region_mask)

        return overlap / previous_area

    def compute_dln_from_tracks(
        self,
        frame: int,
        previous_box_dimensions_dln: Optional[BoxDimensionsDln] = None,
        additional_tracks: Optional[list[CellTrack]] = None,
        relative: bool = True,
    ) -> BoxDimensionsDln:
        """
        Compute Delaunay triangulation at given frame.

        Parameters
        ----------
        frame : int
            Frame at which Delaunay triangulation is computed.
        previous_box_dimensions_dln : Optional[BoxDimensionsDln] = None
            If specified and current track has no spot at frame, no computation is done
            and previous_box_dimensions_dln is returned.
        additional_tracks : Optional[list[CellTrack]] = None
            If specified, perform computation on both current track and these additional tracks.
        relative : bool = True
            If True, indexes are relative to current track position.

        Returns
        -------
        BoxDimensionsDln : Track(s) Delaunay triangulation.

        """
        tracks = [self]
        if additional_tracks is not None:
            tracks = tracks + additional_tracks

        box_dimensions_dln, track_frame_points = get_whole_box_dimensions_dln(
            tracks, frame
        )

        # If missing spot at this frame...
        if box_dimensions_dln.is_empty():
            # ... use previous frame data if provided
            if previous_box_dimensions_dln:
                return previous_box_dimensions_dln

            # ... or try with previous frame if not
            for _ in range(CellTrack.max_frame_gap):
                frame = frame - 1
                (
                    box_dimensions_dln,
                    track_frame_points,
                ) = get_whole_box_dimensions_dln(tracks, frame)
                if not box_dimensions_dln.is_empty():
                    break

        # Should not be empty after this loop
        if box_dimensions_dln.is_empty():
            raise ValueError(
                f"No previous dln & Tracks with no spots in {CellTrack.max_frame_gap} frames in a row"
            )

        # Else, compute convex hull and Delaunay triangulation
        # Switch dimensions
        if relative:
            track_frame_points = [
                [y - box_dimensions_dln.min_y, x - box_dimensions_dln.min_x]
                for x, y in track_frame_points
            ]
        else:
            track_frame_points = [[y, x] for x, y in track_frame_points]
        # Compute hull
        hull = ConvexHull(points=track_frame_points)
        box_dimensions_dln.dln = Delaunay(
            np.array(track_frame_points)[hull.vertices]
        )

        return box_dimensions_dln

    def get_spots_data(
        self, raw_spots: list[CellSpot], raw_video: np.array
    ) -> list[np.array]:
        """
        Generate crops around cells for metaphase CNN inference.

        Parameters
        ----------
        raw_spots : list[CellSpot]
            All video spots.
        raw_video : np.array
            TYXC

        Returns
        -------
        list[np.array]: Cell crops (CYX)

        """

        spot_abs_positions = {}  # {frame: BoxDimensions}
        cell_crops = []  # CYX

        for spot in raw_spots:
            # Ignore spots before or after current track
            if spot.frame < self.start or spot.frame > self.stop:
                continue
            # Ignore spots not in current track
            if spot.id not in self.track_spots_ids:
                continue
            # Store positions
            spot_abs_positions[spot.frame] = BoxDimensions(
                spot.abs_min_x,
                spot.abs_max_x,
                spot.abs_min_y,
                spot.abs_max_y,
            )
            # Store all spots
            spot.track_id = self.track_id  # add track information
            self.spots[spot.frame] = spot

        # If no spot in track for current frame, use previous frame position
        for frame in range(self.start, self.stop + 1):
            if frame in spot_abs_positions:
                min_y, max_y, min_x, max_x = (
                    spot_abs_positions[frame].min_y,
                    spot_abs_positions[frame].max_y,
                    spot_abs_positions[frame].min_x,
                    spot_abs_positions[frame].max_x,
                )
            nucleus = raw_video[frame, min_y:max_y, min_x:max_x, :]  # YXC
            nucleus = np.moveaxis(nucleus, -1, 0)  # CYX
            cell_crops.append(nucleus)

        self.number_spots = len(cell_crops)
        return cell_crops

    @staticmethod
    def generate_tracks_from_spots(
        spots: dict[int, list[CellSpot]],
        linking_max_distance: int,
        gap_closing_max_distance: int,
    ) -> list[CellTrack]:
        """
        Generate tracks from spots.
        """
        max_frame_gap = CellTrack.max_frame_gap
        raise NotImplementedError



import os
import pickle
from typing import Optional

from ..factories.results_saving_factory import ResultsSavingFactory
from ..utils.mitosis_track import MitosisTrack


def perform_results_saving(
    exported_mitoses_dir: str,
    show: bool = False,
    save_dir: Optional[str] = None,
    verbose: bool = False,
) -> None:
    """
    Perform a series of tests, prints and plots following process.
    """
    # Create save_dir if specified and it does not exist
    if save_dir is not None and not os.path.exists(save_dir):
        os.makedirs(save_dir)

    mitosis_tracks: list[MitosisTrack] = []
    # Iterate over "bin" files in exported_mitoses_dir
    for state_path in os.listdir(exported_mitoses_dir):
        with open(os.path.join(exported_mitoses_dir, state_path), "rb") as f:
            mitosis_track: MitosisTrack = pickle.load(f)
            mitosis_track.adapt_deprecated_attributes()
            mitosis_tracks.append(mitosis_track)

    # Define lists and dictionaries to store results
    results_saving_factory = ResultsSavingFactory()
    results_saving_factory.update_cut_times(mitosis_tracks, verbose)

    # Protect against no detection
    if len(results_saving_factory.first_cut_times) == 0:
        print("No mitosis detected.")
        return

    # Perform a series of tests, prints and plots
    results_saving_factory.perform_t_test()
    results_saving_factory.print_analysis_summary(mitosis_tracks)
    results_saving_factory.save_csv_results(mitosis_tracks, save_dir)
    results_saving_factory.box_plot_cut_differences(show, save_dir)
    results_saving_factory.plot_cut_distributions(show, save_dir)



import os
import pickle
from typing import Optional

from cut_detector.data.tools import get_data_path
from cut_detector.utils.trackmate_track import TrackMateTrack
from cut_detector.utils.trackmate_spot import TrackMateSpot


def load_tracks_and_spots(
    trackmate_tracks_path: str, spots_path: str
) -> tuple[list[TrackMateTrack], list[TrackMateSpot]]:
    """
    Load saved spots and tracks generated from Trackmate xml file.
    """
    trackmate_tracks: list[TrackMateTrack] = []
    for track_file in os.listdir(trackmate_tracks_path):
        with open(os.path.join(trackmate_tracks_path, track_file), "rb") as f:
            trackmate_track: TrackMateTrack = pickle.load(f)
            trackmate_track.adapt_deprecated_attributes()
            trackmate_tracks.append(trackmate_track)

    spots: list[TrackMateSpot] = []
    for spot_file in os.listdir(spots_path):
        with open(os.path.join(spots_path, spot_file), "rb") as f:
            spots.append(pickle.load(f))

    return trackmate_tracks, spots


def main(
    segmentation_results_path: Optional[str] = os.path.join(
        get_data_path("segmentation_results"), "example_video.bin"
    ),
    trackmate_tracks_path: Optional[str] = os.path.join(
        get_data_path("tracks"), "example_video"
    ),
    spots_path: Optional[str] = os.path.join(
        get_data_path("spots"), "example_video"
    ),
):
    # Load Cellpose results
    with open(segmentation_results_path, "rb") as f:
        cellpose_results = pickle.load(f)

    # TODO: create spots from Cellpose results
    # TODO: perform tracking using laptrack

    # Load TrackMate results to compare... make sure they match!
    trackmate_tracks, trackmate_spots = load_tracks_and_spots(
        trackmate_tracks_path, spots_path
    )


if __name__ == "__main__":
    main()



import os
from typing import Optional
import pickle
import numpy as np
from matplotlib import pyplot as plt

from cnn_framework.utils.readers.tiff_reader import TiffReader

from cut_detector.data.tools import get_data_path
from cut_detector.models.tools import get_model_path
from cut_detector.factories.mt_cut_detection_factory import (
    MtCutDetectionFactory,
)
from cut_detector.utils.bridges_classification.template_type import (
    TemplateType,
)
from cut_detector.utils.mitosis_track import MitosisTrack


def re_organize_channels(image):
    """
    From any order to TYXC.
    """
    image = image.squeeze()
    assert image.ndim == 4, "Image must be 4D: time, channels, height, width"

    # Get dimension with minimum size and put last
    min_dim = np.argmin(image.shape)
    image = np.moveaxis(image, min_dim, -1)

    # Get dimension with minimum size in first 3 dimensions and put first
    min_dim = np.argmin(image.shape[:3])
    image = np.moveaxis(image, min_dim, 0)

    return image


def main(
    image_path: Optional[str] = get_data_path("videos"),
    mitosis_path: Optional[str] = get_data_path("mitoses"),
    hmm_bridges_parameters_file: Optional[str] = os.path.join(
        get_model_path("hmm"), "hmm_bridges_parameters.npz"
    ),
    bridges_mt_cnn_model_path: Optional[str] = get_model_path(
        "bridges_mt_cnn"
    ),
    display_predictions_analysis=True,
    display_crops=True,
):
    # If paths are directories, take their first file
    if os.path.isdir(image_path):
        image_path = os.path.join(image_path, os.listdir(image_path)[0])
    if os.path.isdir(mitosis_path):
        mitosis_path = os.path.join(mitosis_path, os.listdir(mitosis_path)[0])

    # Read data: image, mitosis_track
    image = TiffReader(image_path, respect_initial_type=True).image
    video = re_organize_channels(image)  # TYXC
    with open(mitosis_path, "rb") as f:
        mitosis_track: MitosisTrack = pickle.load(f)
        mitosis_track.adapt_deprecated_attributes()

    template_type = TemplateType.AVERAGE_CIRCLE
    factory = MtCutDetectionFactory(
        template_type=template_type,
    )

    results = factory.update_mt_cut_detection(
        [mitosis_track],
        video,
        hmm_bridges_parameters_file,
        bridges_mt_cnn_model_path,
        debug_mode=True,
    )

    if display_predictions_analysis:
        plt.plot(results["list_class_bridges"][mitosis_track.id])
        if (
            mitosis_track.id in results["list_class_bridges_after_hmm"]
        ):  # if classification possible
            plt.plot(results["list_class_bridges_after_hmm"][mitosis_track.id])
        plt.title("Class bridges")
        plt.show()

    if display_crops:
        # Display series of crops
        for crop in results["crops"][mitosis_track.id]:
            plt.figure()
            plt.imshow(crop[0], cmap="gray")
            plt.show()


if __name__ == "__main__":
    FOLDER_MITOSIS = r"C:\Users\thoma\data\Data Nathalie\mitoses"
    FOLDER_VIDEO = r"C:\Users\thoma\data\Data Nathalie\videos"
    for mitosis_file in os.listdir(FOLDER_MITOSIS):
        local_mitosis_path = os.path.join(FOLDER_MITOSIS, mitosis_file)
        image_name = mitosis_file.split("_mitosis")[0]
        local_image_path = os.path.join(FOLDER_VIDEO, image_name + ".tif")
        main(local_image_path, local_mitosis_path)

    # main()



from .cell_track import CellTrack


class TrackMateTrack(CellTrack):
    """
    Parse TrackMate track from xml file.
    """

    def __init__(self, trackmate_track):
        track_id = int(trackmate_track["@TRACK_ID"])

        track_spots_ids: set[int] = set()

        if isinstance(trackmate_track["Edge"], dict):  # only one edge
            track_spots_ids.add(
                int(trackmate_track["Edge"]["@SPOT_SOURCE_ID"])
            )
            track_spots_ids.add(
                int(trackmate_track["Edge"]["@SPOT_TARGET_ID"])
            )
        else:
            for edge in trackmate_track["Edge"]:
                track_spots_ids.add(int(edge["@SPOT_SOURCE_ID"]))
                track_spots_ids.add(int(edge["@SPOT_TARGET_ID"]))

        start = int(float(trackmate_track["@TRACK_START"]))
        stop = int(float(trackmate_track["@TRACK_STOP"]))

        super().__init__(track_id, track_spots_ids, start, stop)

    def adapt_deprecated_attributes(self) -> None:
        """
        Adapt deprecated attributes.

        Returns
        -------
        None.

        """
        if hasattr(self, "track_spots"):
            self.spots = self.track_spots
            del self.track_spots



from matplotlib import pyplot as plt
import numpy as np
import torch
from torch.utils.data import DataLoader

import cv2
from scipy import ndimage

from cnn_framework.utils.metrics.abstract_metric import AbstractMetric
from cnn_framework.utils.data_sets.dataset_output import DatasetOutput
from cnn_framework.utils.model_managers.cnn_model_manager import (
    CnnModelManager,
)

from .micro_tubules_augmentation import (
    MicroTubulesAugmentation,
)


def transform_image_orientation(image, margin=10, percentile=75):
    """
    Compute image orientation and rotate it to have the bridges horizontal.
    """
    keep_3_channels = False
    if len(image.shape) == 3:
        assert image.shape[0] == 1, "Only one channel image is supported"
        image = image[0]
        keep_3_channels = True

    original_height, original_width = image.shape

    # Select sub-image (containing information)
    sub_image = image[
        original_height // 2 - margin : original_height // 2 + margin,
        original_width // 2 - margin : original_width // 2 + margin,
    ]
    threshold = np.percentile(sub_image, percentile)
    binary_image = (sub_image > threshold).astype(np.uint8)

    # Compute image moments
    moments = cv2.moments(binary_image)
    # Compute orientation
    orientation = 0.5 * np.arctan2(
        2 * moments["mu11"], moments["mu20"] - moments["mu02"]
    )
    # Compute orientation in degrees
    orientation_degrees = np.degrees(orientation)

    # Rotate image and keep original size
    rotated_image = ndimage.rotate(
        image, orientation_degrees, cval=image.min()
    )
    height, width = rotated_image.shape
    rotated_image = rotated_image[
        (height - original_height) // 2 : (height - original_height) // 2
        + original_height,
        (width - original_width) // 2 : (width - original_width) // 2
        + original_width,
    ]

    # Expand first dim
    if keep_3_channels:
        rotated_image = np.expand_dims(rotated_image, axis=0)

    return rotated_image, binary_image


class BridgesMtModelManager(CnnModelManager):
    """
    Model manager for Binary Bridges CNN classification models.
    """

    def augment_image(self, input_tensor: torch.Tensor) -> torch.Tensor:
        """
        Function to augment image following binary bridges principle.
        """
        input_arrays = input_tensor.detach().cpu().numpy()  # BCYX

        augmentation_tool = MicroTubulesAugmentation()
        augmentation_tool.augmentations.pop("top")
        augmentation_tool.augmentations.pop("bottom")

        augmented_array = []
        for orig_input_array in input_arrays:
            # Put first dimension last
            input_array, _ = transform_image_orientation(orig_input_array)
            augmented_images = augmentation_tool.generate_augmentations(
                input_array
            )
            for augmented_dict in augmented_images.values():
                # Put last dimension first
                augmented_image = augmented_dict["image"]  # YXC
                augmented_image = augmented_image.transpose(2, 0, 1)  # CYX
                augmented_array.append(augmented_image)

        augmented_array = np.array(augmented_array)
        augmented_tensor = torch.from_numpy(augmented_array)
        return augmented_tensor

    def model_prediction(
        self,
        dl_element: DatasetOutput,
        dl_metric: AbstractMetric,
        data_loader: DataLoader,
    ) -> None:
        """
        Function to generate outputs from inputs for given model.
        """
        augmented_input = self.augment_image(dl_element.input)
        augmentation_nb = augmented_input.shape[0] // dl_element.input.shape[0]

        dl_element.to_device(self.device)
        augmented_input = augmented_input.to(self.device)

        predictions = torch.softmax(
            self.model(augmented_input.float()), dim=-1
        )

        # From MT detection to cut detection
        predictions = predictions.view(
            -1, augmentation_nb // 2, 2, 2
        )  # augmentation_nb images, 2 scores
        final_predictions = torch.zeros_like(dl_element.target)
        for idx, image_predictions in enumerate(predictions):
            for pair_predictions in image_predictions:
                # NB: no cut means 2 MT
                no_cut_prediction = (
                    pair_predictions[0][1] * pair_predictions[1][1]
                )
                cut_prediction = (
                    pair_predictions[0][1] * pair_predictions[1][0]
                    + pair_predictions[0][0] * pair_predictions[1][1]
                )
                two_cuts_prediction = (
                    pair_predictions[0][0] * pair_predictions[1][0]
                )
                final_predictions[idx][0] += no_cut_prediction
                final_predictions[idx][1] += cut_prediction
                final_predictions[idx][2] += two_cuts_prediction
        dl_element.prediction = final_predictions

        # # Debug display
        # target_argmax = torch.argmax(dl_element.target, dim=1)
        # predictions_argmax = torch.argmax(final_predictions, dim=1)
        # predictions_to_print = predictions.view(-1, augmentation_nb, 2)
        # for img_idx, original_image in enumerate(dl_element.input):
        #     if predictions_argmax[img_idx] == target_argmax[img_idx]:
        #         continue
        #     plt.subplot(1, 1 + augmentation_nb, 1)
        #     mat_original = original_image[0].squeeze().detach().cpu().numpy()
        #     mat_original, _ = transform_image_orientation(mat_original)
        #     plt.title(
        #         f"{self.file_name_encoder.decode(dl_element.encoded_file_name[img_idx])} \n Pred {2 - predictions_argmax[img_idx].detach().cpu().numpy()} MT vs Target {2 - target_argmax[img_idx].detach().cpu().numpy()} MT"
        #     )
        #     plt.imshow(mat_original, cmap="gray")
        #     for plt_idx, idx in enumerate(
        #         range(
        #             augmentation_nb * img_idx, augmentation_nb * (img_idx + 1)
        #         )
        #     ):
        #         plt.subplot(1, 1 + augmentation_nb, 2 + plt_idx)
        #         mat_sub_image = (
        #             augmented_input[idx][0].squeeze().detach().cpu().numpy()
        #         )
        #         augment_pred = (
        #             predictions_to_print[img_idx][plt_idx]
        #             .detach()
        #             .cpu()
        #             .numpy()
        #         )
        #         predicted_category = ["No MT", "MT"][np.argmax(augment_pred)]
        #         plt.title(f"{predicted_category} {str(max(augment_pred))[:4]}")
        #         plt.imshow(mat_sub_image, cmap="gray")

        #     # Display on whole screen
        #     figManager = plt.get_current_fig_manager()
        #     figManager.window.showMaximized()
        #     plt.show()

        # Update metric
        dl_metric.update(
            final_predictions,
            dl_element.target,
            adds=dl_element.additional,
            mean_std=data_loader.dataset.mean_std,
        )



import os
from typing import Optional
from scipy.stats import ttest_1samp
import numpy as np
import matplotlib.pyplot as plt

from ..utils.mitosis_track import MitosisTrack
from ..utils.bridges_classification.impossible_detection import (
    ImpossibleDetection,
)


class ResultsSavingFactory:
    """Factory to save results."""

    def __init__(
        self,
        time_resolution: Optional[int] = 10,
        max_frame: Optional[int] = np.inf,
    ):
        self.time_resolution = time_resolution
        self.max_frame = max_frame  # for debug, 48*6 <-> 48h * 6 frames/hour

        self.first_cut_times: list[int] = []  # time in absolute time
        self.first_cut_times_gt: list[int] = []  # time in absolute time
        self.cut_differences: list[int] = []

        self.mitosis_results_summary: dict[int, int] = {}
        self.gt_mitosis_results_summary: dict[int, int] = {}

        self.p_value = None

    def _print_weird_mitoses(
        self, selected_tracks: list[MitosisTrack], min_acceptable_frame: int
    ) -> None:
        """
        Print useful information about weird mitoses.
        """
        ordered_tracks = [
            track
            for track in selected_tracks
            if track.key_events_frame["first_mt_cut"]
            - track.key_events_frame["cytokinesis"]
            <= min_acceptable_frame
        ]
        ordered_tracks.sort(
            key=lambda x: x.key_events_frame["first_mt_cut"]
            - x.key_events_frame["cytokinesis"]
        )

        print("Weird mitoses (early cut):")
        for mitosis_track in ordered_tracks:
            print("")
            print(
                f"Track: {mitosis_track.id}_{mitosis_track.mother_track_id}_to_{','.join(str(daughter) for daughter in mitosis_track.daughter_track_ids)}"
            )
            print(f"Key events frame: {mitosis_track.key_events_frame}")
            print(
                "video frame cut",
                mitosis_track.key_events_frame["first_mt_cut"]
                - mitosis_track.min_frame
                + 1,
            )
            if mitosis_track.gt_key_events_frame is not None:
                print(
                    f"GT key events frame: {mitosis_track.gt_key_events_frame}"
                )
                print(
                    "video frame cut",
                    mitosis_track.gt_key_events_frame["first_mt_cut"]
                    - mitosis_track.min_frame
                    + 1,
                )

    def update_cut_times(
        self,
        mitosis_tracks: list[MitosisTrack],
        verbose: bool,
        min_acceptable_frame=13,
    ) -> None:
        """Update cut times and summary."""

        selected_tracks: list[MitosisTrack] = []  # kept tracks

        for mitosis_track in mitosis_tracks:
            # Get first cut frame and start of cytokinesis frame
            cyto_frame = mitosis_track.key_events_frame["cytokinesis"]
            cut_frame = mitosis_track.key_events_frame["first_mt_cut"]

            if cut_frame < 0 or cut_frame > self.max_frame:
                # No cut detected, for some reason
                cut_time = None
            else:
                # At least one cut was actually detected
                assert cut_frame >= cyto_frame  # should not be possible
                # Add the difference time to the list
                cut_time = (cut_frame - cyto_frame) * self.time_resolution
                self.first_cut_times.append(cut_time)
                selected_tracks.append(mitosis_track)

            # Update summary
            cut_id = min(cut_frame, 0)  # 0 if normal, negative otherwise

            if cut_id not in self.mitosis_results_summary:
                self.mitosis_results_summary[cut_id] = 0
            self.mitosis_results_summary[cut_id] += 1

            # Ignore if information is not there
            if (
                mitosis_track.gt_key_events_frame is None
                or "first_mt_cut" not in mitosis_track.gt_key_events_frame
            ):
                continue

            if (
                mitosis_track.gt_key_events_frame["first_mt_cut"]
                > self.max_frame
            ):
                # Ignore mitoses that are too late
                continue

            cut_time_gt = (
                mitosis_track.gt_key_events_frame["first_mt_cut"]
                - mitosis_track.gt_key_events_frame["cytokinesis"]
            ) * self.time_resolution
            self.first_cut_times_gt.append(cut_time_gt)

            if cut_time is not None:
                self.cut_differences.append(cut_time - cut_time_gt)

            # Update ground truth summary
            if cut_id not in self.gt_mitosis_results_summary:
                self.gt_mitosis_results_summary[cut_id] = 0
            self.gt_mitosis_results_summary[cut_id] += 1

        # Display few details on mitoses with very early cut
        if verbose:
            self._print_weird_mitoses(selected_tracks, min_acceptable_frame)

    def perform_t_test(self, alpha=0.05) -> None:
        """
        Compute t-test of the differences, which is supposed to be 0.
        """
        if len(self.cut_differences) == 0:
            return None

        t_result = ttest_1samp(self.cut_differences, 0)
        self.p_value = t_result.pvalue

        # Output the results
        print(f"P-value: {self.p_value}")

        # Check if the result is statistically significant (e.g., using a significance level of 0.05)
        if self.p_value < alpha:
            print(
                "Reject the null hypothesis: There is a statistically significant difference."
            )
        else:
            print(
                "Fail to reject the null hypothesis: There is no statistically significant difference."
            )

    def print_analysis_summary(self, mitosis_tracks) -> None:
        """Print analysis summary."""
        print(
            f"\n Among the {len(mitosis_tracks)} mitoses detected, there are:"
        )
        for cut_id, count in self.mitosis_results_summary.items():
            print(
                f"    - {count} mitoses in category {ImpossibleDetection(cut_id).name}"
            )

        if len(self.first_cut_times_gt) == 0:
            return
        print(
            f"\n Among the {len(self.first_cut_times_gt)} mitoses annotated, there are:"
        )
        for cut_id, count in self.gt_mitosis_results_summary.items():
            print(
                f"    - {count} mitoses in category {ImpossibleDetection(cut_id).name}"
            )

    def box_plot_cut_differences(
        self, show: bool, save_dir: Optional[str]
    ) -> None:
        """
        Plot box plot of cut differences
        """
        if len(self.cut_differences) == 0:
            return

        plt.figure()
        plt.boxplot(self.cut_differences)
        plt.title("Box plot of cut differences")
        plt.ylabel("Cut difference (min)")

        if show:
            plt.show()

        if save_dir is not None:
            plt.savefig(os.path.join(save_dir, "first_mt_cut_differences.png"))

    def plot_cut_distributions(
        self,
        show: bool,
        save_dir: Optional[str],
    ) -> None:
        """
        Plot distribution of cut differences
        """

        plt.figure()

        for cut_times, name in zip(
            [self.first_cut_times, self.first_cut_times_gt],
            ["Predicted", "Ground truth"],
        ):
            if len(cut_times) == 0:  # ignore empty gt list
                continue
            curve_values = []
            nb_values = len(cut_times)
            number_of_cut = 0
            for i in range(max(cut_times) + 1):
                count = cut_times.count(i)
                number_of_cut += count
                curve_values.append(number_of_cut / nb_values)
            plt.plot(curve_values, label=name)

        # Quartiles and median
        predicted_median = np.median(self.first_cut_times)
        gt_median = (
            np.median(self.first_cut_times_gt)
            if len(self.first_cut_times_gt)
            else None
        )
        predicted_q1 = np.percentile(self.first_cut_times, 25)
        predicted_q3 = np.percentile(self.first_cut_times, 75)
        plt.axvline(
            x=predicted_q1,
            color="r",
            linestyle="--",
            linewidth=0.5,
            label=f"Detected Q1: {predicted_q1}",
        )
        plt.axvline(
            x=predicted_median,
            color="r",
            linestyle="-",
            linewidth=0.5,
            label=f"Detected median: {predicted_median}"
            + (f"vs {gt_median} Ground truth median" if gt_median else ""),
        )
        plt.axvline(
            x=predicted_q3,
            color="r",
            linestyle="--",
            linewidth=0.5,
            label=f"Detected Q3: {predicted_q3}",
        )

        # add t test result to legend
        if self.p_value is not None:
            plt.legend(
                loc="lower right", title=f"t test p-value: {self.p_value:.2f}"
            )
        else:
            plt.legend(loc="lower right")

        plt.title(
            f"Proportion of MT cuts over time - {len(self.first_cut_times)} predicted mitoses"
            + (
                f"- {len(self.first_cut_times_gt)} ground truth mitoses"
                if len(self.first_cut_times_gt)
                else ""
            )
        )
        plt.xlabel("Time (min)")
        plt.ylabel("Proportion of first cut")

        if show:
            plt.show()

        if save_dir is not None:
            plt.savefig(
                os.path.join(save_dir, "first_mt_cut_distributions.png")
            )

    def save_csv_results(
        self,
        mitosis_tracks: list[MitosisTrack],
        save_dir: Optional[str] = None,
    ) -> None:
        """Save results in CSV file."""
        if save_dir is None:
            return

        csv_path = os.path.join(save_dir, "results.csv")
        # Create CSV results file if it does not exist
        if not os.path.exists(csv_path):
            with open(csv_path, "w") as f:
                f.write(
                    "id;mother track;daughter track(s);metaphase frame;cytokinesis frame;first MT cut frame;second MT cut frame;first MT cut time;second MT cut time\n"
                )
            f.close()

        # Store useful results in global results file
        with open(csv_path, "a") as f:
            for mitosis_track in mitosis_tracks:
                f.write(f"{mitosis_track.id};")
                f.write(f"{mitosis_track.mother_track_id};")
                daughter_ids = ",".join(
                    [str(d) for d in mitosis_track.daughter_track_ids]
                )
                f.write(f"{daughter_ids};")
                f.write(f"{mitosis_track.key_events_frame['metaphase']};")
                cytokinesis_frame = mitosis_track.key_events_frame[
                    "cytokinesis"
                ]
                f.write(f"{cytokinesis_frame};")
                first_cut_frame = mitosis_track.key_events_frame[
                    "first_mt_cut"
                ]
                f.write(f"{first_cut_frame};")
                second_cut_frame = mitosis_track.key_events_frame[
                    "second_mt_cut"
                ]
                f.write(f"{second_cut_frame};")
                first_cut_time = (
                    (first_cut_frame - cytokinesis_frame)
                    * self.time_resolution
                    if first_cut_frame >= 0
                    else ""
                )
                f.write(f"{first_cut_time};")
                second_cut_time = (
                    (second_cut_frame - cytokinesis_frame)
                    * self.time_resolution
                    if second_cut_frame >= 0
                    else ""
                )
                f.write(f"{second_cut_time};\n")
        f.close()



from __future__ import annotations

from copy import deepcopy
import json
from typing import Optional, Tuple, Union
import numpy as np
from skimage.morphology import extrema, opening
import xmltodict
from munch import Munch
from scipy import ndimage

from ..constants.annotations import (
    NAMES_DICTIONARY,
    get_class_ids_after_first_mt_cut,
    get_class_ids_after_second_mt_cut,
    get_class_ids_after_first_membrane_cut,
)
from ..constants.tracking import (
    CYTOKINESIS_DURATION,
    FRAMES_AROUND_METAPHASE,
    METAPHASE_INDEX,
    MINIMUM_DISTANCE_TO_BORDER,
)
from .mid_body_spot import MidBodySpot
from .trackmate_track import TrackMateTrack
from .box_dimensions_dln import BoxDimensionsDln
from .box_dimensions import BoxDimensions
from .bridges_classification.impossible_detection import ImpossibleDetection
from .image_tools import resize_image, smart_cropping
from .tools import cell_counter_frame_to_video_frame


class MitosisTrack:
    """
    A class to store the information of a mitosis track.
    """

    def __init__(
        self,
        mother_track_id: int,
        daughter_track_id: int,
        metaphase_frame: int,
    ):
        # Elementary information
        self.mother_track_id = mother_track_id
        self.daughter_track_ids = [daughter_track_id]
        self.id: Optional[int] = None

        self.metaphase_frame = metaphase_frame

        # Key events: metaphase/cytokinesis/first_mt_cut/second_mt_cut/first_membrane_cut
        # Absolute frame
        self.key_events_frame: dict[str, Union[int, ImpossibleDetection]] = {}
        self.gt_key_events_frame: Optional[dict[str, int]] = None

        # Time
        self.min_frame: Optional[int] = None
        self.max_frame: Optional[int] = None

        # Position
        self.position = BoxDimensions()

        # Delaunay triangulation, by frame
        self.dln_positions: dict[int, BoxDimensionsDln] = {}

        # Used for matching between ground truth and prediction
        self.matched = False

        # Mid body spot indexed by absolute frame
        self.mid_body_spots: dict[int, MidBodySpot] = {}
        self.gt_mid_body_spots: Optional[dict[int, MidBodySpot]] = None

        # Used to know if the track is near the border of the video
        self.is_near_border = False

    def is_same_mitosis(
        self, mother_track_id: int, metaphase_frame: int
    ) -> bool:
        return (
            self.mother_track_id == mother_track_id
            and self.metaphase_frame == metaphase_frame
        )

    def add_daughter_track(self, daughter_track_id: int) -> None:
        self.daughter_track_ids.append(daughter_track_id)

    def get_mother_daughters_tracks(
        self, tracks: list[TrackMateTrack]
    ) -> Tuple[TrackMateTrack, list[TrackMateTrack]]:
        mother_track = [
            track for track in tracks if track.track_id == self.mother_track_id
        ][0]
        daughter_tracks = [
            track
            for track in tracks
            if track.track_id in self.daughter_track_ids
        ]
        return mother_track, daughter_tracks

    def _add_dln_position(
        self, frame: int, frame_dimensions: BoxDimensionsDln
    ) -> None:
        self.dln_positions[frame] = deepcopy(frame_dimensions)
        # Update absolute min and max accordingly
        self.position.update_from_box_dimensions(frame_dimensions)

    def update_mitosis_start_end(
        self,
        trackmate_tracks: list[TrackMateTrack],
        mitosis_tracks: list[MitosisTrack],
    ) -> None:
        # Get all tracks involved in current mitosis
        mother_track, daughter_tracks = self.get_mother_daughters_tracks(
            trackmate_tracks
        )

        # Get min and max frame of current mitosis
        # Min is the metaphase frame minus FRAMES_AROUND_METAPHASE, protected against frames before start of mother track
        min_frame = max(
            mother_track.start,
            self.metaphase_frame - FRAMES_AROUND_METAPHASE,
        )
        # For each daughter track, the end is the end of the track OR the next metaphase event of this track
        max_frame = mother_track.stop
        for track in [mother_track] + daughter_tracks:
            track_end_frame = track.stop
            for track_to_merge_bis in mitosis_tracks:
                if (
                    track_to_merge_bis.mother_track_id == track.track_id
                    and track_to_merge_bis.metaphase_frame
                    > self.metaphase_frame  # other mitosis should be strictly after
                ):
                    track_end_frame = min(
                        track_end_frame, track_to_merge_bis.metaphase_frame
                    )
            max_frame = min(max_frame, track_end_frame)

        # Update mitosis_track
        self.min_frame = min_frame
        self.max_frame = max_frame

    def update_is_near_border(self, raw_video: np.array) -> None:
        """
        Parameters
        ----------
        raw_video: np.array
            TYXC

        """

        max_height, max_width = raw_video.shape[1], raw_video.shape[2]

        cyto_frame = self.key_events_frame["cytokinesis"]
        last_frame = cyto_frame + CYTOKINESIS_DURATION

        # get mitosis coordinates between cyto_frame and last_frame
        min_dist_to_border = np.inf
        for frame in range(cyto_frame, last_frame + 1):
            if frame not in self.mid_body_spots:
                continue

            # get mid-body coordinates
            mid_body_frame = self.mid_body_spots[frame]
            x_rel = mid_body_frame.x
            y_rel = mid_body_frame.y

            x_abs = x_rel + self.position.min_x
            y_abs = y_rel + self.position.min_y
            mid_body_coordinates = (x_abs, y_abs)

            # get distance to border
            min_x = min(
                mid_body_coordinates[0], max_width - mid_body_coordinates[0]
            )
            min_y = min(
                mid_body_coordinates[1], max_height - mid_body_coordinates[1]
            )

            min_dist_to_border = min(min_dist_to_border, min_x, min_y)

        self.is_near_border = min_dist_to_border < MINIMUM_DISTANCE_TO_BORDER

    def update_key_events_frame(
        self, trackmate_tracks: list[TrackMateTrack]
    ) -> None:
        # Get all tracks involved in current mitosis
        mother_track, daughter_tracks = self.get_mother_daughters_tracks(
            trackmate_tracks
        )

        # Store first metaphase frame
        for frame in range(self.metaphase_frame, mother_track.start, -1):
            # Some frames may be missing since gap closing is allowed
            if frame not in mother_track.spots:
                continue
            if mother_track.spots[frame].predicted_phase != METAPHASE_INDEX:
                self.key_events_frame["metaphase"] = frame + 1
                break

        # If no metaphase frame found, consider it is the first frame of the mother track
        if "metaphase" not in self.key_events_frame:
            self.key_events_frame["metaphase"] = mother_track.start

        # Store first cytokinesis frame - considered as the first frame of daughter tracks
        self.key_events_frame["cytokinesis"] = min(
            [track.start for track in daughter_tracks]
        )

    def update_mitosis_position_dln(
        self, trackmate_tracks: list[TrackMateTrack]
    ) -> None:
        """
        Update positions of mitosis for each frame and Delaunay triangulation
        """

        min_frame, max_frame = self.min_frame, self.max_frame
        mother_track, daughter_tracks = self.get_mother_daughters_tracks(
            trackmate_tracks
        )

        previous_box_dimensions_dln = None
        for frame in range(min_frame, max_frame + 1):
            box_dimensions_dln = mother_track.compute_dln_from_tracks(
                frame,
                previous_box_dimensions_dln,
                additional_tracks=daughter_tracks,
            )
            # Store in case next frame is missing
            previous_box_dimensions_dln = box_dimensions_dln
            # Update accordingly
            self._add_dln_position(frame, box_dimensions_dln)

    def generate_video_movie(
        self, raw_video: np.array
    ) -> Tuple[np.array, np.array]:
        """
        Parameters
        ----------
        raw_video : np.array
            initial video, TYXC

        Returns
        ----------
        mitosis_movie : np.array
            mitosis movie, TYXC
        mask_movie : np.array
            mask movie, TYX

        """

        mitosis_movie, mask_movie = [], []
        for frame in range(self.min_frame, self.max_frame + 1):
            # Get useful data for current frame
            min_x = self.dln_positions[frame].min_x
            max_x = self.dln_positions[frame].max_x
            min_y = self.dln_positions[frame].min_y
            max_y = self.dln_positions[frame].max_y
            dln = self.dln_positions[frame].dln

            # Extract frame image, big enough to keep all spots for current track
            frame_image = raw_video[
                frame,
                self.position.min_y : self.position.max_y,
                self.position.min_x : self.position.max_x,
                :,
            ]  # YXC

            # Generate mask with Delaunay triangulation
            current_frame_shape = (
                max_y - min_y,
                max_x - min_x,
            )  # current spot
            indices = np.stack(np.indices(current_frame_shape), axis=-1)
            out_idx = np.nonzero(dln.find_simplex(indices) + 1)
            single_channel_mask = np.zeros(current_frame_shape)
            single_channel_mask[out_idx] = 1

            # Construct mask image
            mask_image = np.stack(
                [single_channel_mask] * raw_video.shape[-1], axis=0
            )  # CYX
            mask_image = resize_image(
                mask_image,
                method="zero",
                pad_margin_h=[
                    min_y - self.position.min_y,
                    self.position.max_y - max_y,
                ],
                pad_margin_w=[
                    min_x - self.position.min_x,
                    self.position.max_x - max_x,
                ],
            )[
                0, ...
            ]  # YX

            mitosis_movie.append(frame_image)
            mask_movie.append(mask_image)

        mitosis_movie = np.array(mitosis_movie)  # TYXC
        mask_movie = np.array(mask_movie)  # TYX

        # If mid-bodies are already computed, add them to the mitosis movie
        if self.mid_body_spots:
            mitosis_movie = self.add_mid_body_movie(mitosis_movie, mask_movie)
            return mitosis_movie[..., :-1], mitosis_movie[..., -1].squeeze()

        return mitosis_movie, mask_movie

    def generate_mitosis_summary(
        self, raw_tracks: list[TrackMateTrack], save_path: str
    ) -> None:
        """
        Unused so far.
        Might be improved with all useful information, saved to csv...
        """
        mitosis_summary = {}

        mother_track, daughter_tracks = self.get_mother_daughters_tracks(
            raw_tracks
        )
        daughters_first_frame = min([track.start for track in daughter_tracks])

        for idx, frame in enumerate(range(self.min_frame, self.max_frame + 1)):
            # Extreme case where mother track is not present at beginning of mitosis movie
            if frame not in mother_track.spots:
                mitosis_summary[idx + 1] = "interphase"
                continue
            # Telophase defined as first frame after metaphase or daughters first frame
            if frame >= self.metaphase_frame or frame >= daughters_first_frame:
                mitosis_summary[idx + 1] = "telophase"
            # Metaphase according to CNN + HMM prediction
            elif mother_track.spots[frame].predicted_phase == METAPHASE_INDEX:
                mitosis_summary[idx + 1] = "metaphase"
            # In other cases, interphase
            else:
                mitosis_summary[idx + 1] = "interphase"

        # Save mitosis summary
        with open(save_path, "w") as f:
            json.dump(mitosis_summary, f)

    def is_possible_match(self, other_track: MitosisTrack) -> bool:
        """
        Check if two tracks are a possible match. Other track is typically a ground truth track.
        Match is possible if there is an overlap between the two tracks,
        and other track starts no earlier/no later than FRAMES_AROUND_METAPHASE around self start.
        """
        if (
            abs(other_track.metaphase_frame - self.metaphase_frame)
            > FRAMES_AROUND_METAPHASE
        ):
            return False

        return self.position.overlaps(other_track.position)

    def add_mid_body_movie(
        self, mitosis_movie: np.array, mask_movie: np.array
    ) -> np.array:
        """
        Parameters
        ----------
        mitosis_movie: TYXC
        mask_movie: TYX

        Returns
        ----------
        spots_video: TYX C=1
        """

        video_shape = mitosis_movie.shape[:3]
        spots_video = np.zeros(video_shape)  # TYX

        for absolute_frame, spot in self.mid_body_spots.items():
            # Create 1 circle around spot position
            square_size = 2
            spots_video[
                absolute_frame - self.min_frame,
                spot.y - square_size : spot.y + square_size,
                spot.x - square_size : spot.x + square_size,
            ] = 1

        # Add empty dimension at end
        mid_body_movie = np.expand_dims(spots_video, axis=-1)  # TYX C=1

        # Mix mid-body and mask movie
        mask_movie = np.expand_dims(mask_movie, axis=-1)  # TYX C=1
        mid_body_mask_movie = mask_movie + mid_body_movie  # TYX C=1

        # Cast mid_body_mask_movie to mitosis_movie dtype
        mid_body_mask_movie = mid_body_mask_movie.astype(mitosis_movie.dtype)

        mitosis_movie = np.concatenate(
            [mitosis_movie, mid_body_mask_movie], axis=-1
        )  # TYX C=C+1

        return mitosis_movie

    def update_mid_body_ground_truth(
        self, annotation_file: str, nb_channels: int
    ) -> None:
        """
        Parameters
        ----------
        annotation_file: .xml file with annotations from CellCounter
        nb_channels: number of channels in mitosis movie (very likely to be 4)

        """

        # Initialize gt_key_events_frame - first two events are shared
        self.gt_key_events_frame = {
            "metaphase": self.key_events_frame["metaphase"],
            "cytokinesis": self.key_events_frame["cytokinesis"],
        }
        self.gt_mid_body_spots = {}

        # Read data
        with open(annotation_file) as fd:
            doc = Munch.fromDict(xmltodict.parse(fd.read()))

        for i, type_data in enumerate(
            doc.CellCounter_Marker_File.Marker_Data.Marker_Type
        ):
            assert i == int(type_data.Type) - 1  # order must be kept
            # Ignore if no data
            if "Marker" not in type_data:
                continue
            markers = type_data.Marker
            if not isinstance(markers, list):
                markers = [markers]
            # Sort markers by frame
            markers = sorted(markers, key=lambda x: int(x.MarkerZ))
            for marker in markers:
                x_pos, y_pos, frame = (
                    int(marker.MarkerX),
                    int(marker.MarkerY),
                    cell_counter_frame_to_video_frame(
                        int(marker.MarkerZ), nb_channels
                    ),
                )
                # Create associated spot
                self.gt_mid_body_spots[frame + self.min_frame] = MidBodySpot(
                    frame, x=x_pos, y=y_pos
                )

            # If Name is missing of wrong, assume it is i
            if (
                "Name" not in type_data
                or type_data.Name not in NAMES_DICTIONARY
            ):
                class_index = i
            else:
                class_index = NAMES_DICTIONARY[type_data.Name]
                assert class_index == i

            class_first_frame = cell_counter_frame_to_video_frame(
                int(markers[0].MarkerZ), nb_channels
            )
            class_abs_first_frame = class_first_frame + self.min_frame

            # First MT cut
            if (
                "first_mt_cut" not in self.gt_key_events_frame
                and class_index in get_class_ids_after_first_mt_cut()
            ):
                assert (
                    class_abs_first_frame
                    >= self.gt_key_events_frame["cytokinesis"]
                )  # after metaphase
                self.gt_key_events_frame["first_mt_cut"] = (
                    class_abs_first_frame
                )

            # Second MT cut
            if (
                "second_mt_cut" not in self.gt_key_events_frame
                and class_index in get_class_ids_after_second_mt_cut()
            ):
                assert (
                    class_abs_first_frame
                    >= self.gt_key_events_frame["first_mt_cut"]
                )  # after first MT cut
                self.gt_key_events_frame["second_mt_cut"] = (
                    class_abs_first_frame
                )

            # First membrane cut
            if (
                "first_membrane_cut" not in self.gt_key_events_frame
                and class_index in get_class_ids_after_first_membrane_cut()
            ):
                assert (
                    class_abs_first_frame
                    >= self.gt_key_events_frame["first_mt_cut"]
                )  # after first MT cut
                self.gt_key_events_frame["first_membrane_cut"] = (
                    class_abs_first_frame
                )

    def evaluate_mid_body_detection(
        self, tolerance=10, percent_seen=0.9
    ) -> bool:
        """
        Mid_body is considered as detected if during at least percent_seen % of frames
        between cytokinesis and second MT cut it is at most tolerance pixels away
        from ground truth during this interval.

        Parameters
        ----------
        tolerance: maximum distance between ground truth and prediction to consider a match
        percent_seen: minimum percentage of frames where mid_body is seen to consider a match
        """

        position_difference = []

        # Check frames until second MT cut or end of annotations
        max_frame = (
            self.gt_key_events_frame["second_mt_cut"]
            if "second_mt_cut" in self.gt_key_events_frame
            else max(self.gt_mid_body_spots.keys())
        )

        for frame in range(self.gt_key_events_frame["cytokinesis"], max_frame):
            if frame not in self.gt_mid_body_spots:
                continue
            if frame not in self.mid_body_spots:
                position_difference.append(1e3)  # random huge value
                continue
            position_difference.append(
                self.gt_mid_body_spots[frame].distance_to(
                    self.mid_body_spots[frame]
                )
            )

        # Get percent_seen th percentile of position difference
        position_difference = np.array(position_difference)
        max_position_difference = np.quantile(
            position_difference, percent_seen
        )

        is_correctly_detected = max_position_difference < tolerance
        position_difference_wo_outliers = np.array(
            [pos for pos in position_difference if pos < 1e3]
        )
        percent_detected = int(
            (
                len(position_difference_wo_outliers)
                / len(position_difference)
                * 100
            )
        )
        average_position_difference = int(
            (
                position_difference_wo_outliers.mean()
                if len(position_difference_wo_outliers) > 0
                else 1e3
            )
        )

        return (
            is_correctly_detected,
            percent_detected,
            average_position_difference,
        )

    def light_spot_detected(
        self,
        video: np.ndarray,
        first_cut_frame: int,
        length_light_spot: int,
        crop_size_light_spot: int,
        h_maxima_light_spot: int,
        intensity_threshold_light_spot: int,
        center_tolerance_light_spot: int,
        min_percentage_light_spot: float,
        print_enabled=False,
    ) -> bool:
        """
        Check if there is a light spot in crops of size crop_size_light_spot
        around the mid-body, in length_light_spot frames around the first micro-tubules cut.

        Spots are detected using h-maxima method with h=h_maxima_light_spot.
        Ignore spots with intensity lower than intensity_threshold_light_spot.
        Ignore spots close to the center (potential mid-bodies), i.e. within
        center_tolerance_light_spot pixels.

        Light spot is considered as detected if at least in min_percentage_light_spot % of frames.
        """
        # Get the mitosis video crop
        video = video[
            :,
            self.position.min_y : self.position.max_y,
            self.position.min_x : self.position.max_x,
            :,
        ]

        nb_spot_detected = 0
        frame_counted = 0
        # Iterate over video frames
        for i in range(-length_light_spot // 2, length_light_spot // 2):
            frame = first_cut_frame + i

            # Make sure mid-body exists at frame
            if not self.mid_body_spots or frame not in self.mid_body_spots:
                continue

            # Get mid-body coordinates
            mid_body_frame = self.mid_body_spots[frame]
            x_pos, y_pos = mid_body_frame.x, mid_body_frame.y

            # Extract image and crop on the midbody
            img = np.transpose(video[frame, ...], (2, 0, 1))  # CYX
            crop = smart_cropping(
                img, crop_size_light_spot, x_pos, y_pos, pad=True
            )[
                0, ...
            ]  # YX

            # Perform opening to remove small spots and apply h_maxima to get potential spots
            filtered_image = opening(crop, footprint=np.ones((3, 3)))
            local_maxima = extrema.h_maxima(
                filtered_image, h_maxima_light_spot
            )

            # Label spot regions and remove inconsistent ones
            labeled_local_maxima, nb_labels = ndimage.label(
                local_maxima, structure=np.ones((3, 3))
            )
            for label in range(1, nb_labels + 1):
                # Labels intensity in original image has to be higher than threshold
                if (
                    crop[np.where(labeled_local_maxima == label)].mean()
                    < intensity_threshold_light_spot
                ):
                    labeled_local_maxima[labeled_local_maxima == label] = 0
            # Re-label accordingly
            labeled_local_maxima, nb_labels = ndimage.label(
                labeled_local_maxima > 0, structure=np.ones((3, 3))
            )

            # Get center of mass to locate spots
            spots = ndimage.center_of_mass(
                local_maxima, labeled_local_maxima, range(1, nb_labels + 1)
            )
            spots = np.asarray(spots, dtype=np.int64)

            # Remove spots that are too close to the center
            for spot in spots:
                if (
                    np.abs(spot[0] - crop_size_light_spot)
                    < center_tolerance_light_spot
                ) and (
                    np.abs(spot[1] - crop_size_light_spot)
                    < center_tolerance_light_spot
                ):
                    spots = np.delete(
                        spots, np.where((spots == spot).all(axis=1))[0], axis=0
                    )

            if len(spots) > 0:
                nb_spot_detected += 1
            frame_counted += 1

        # Light spot is considered as detected if at least in MIN_PERCENTAGE_LIGHTSPOT % of frames
        if frame_counted > 0:
            percentage_spot_detected = nb_spot_detected / frame_counted
            spot_detected = (
                percentage_spot_detected >= min_percentage_light_spot
            )
        else:
            spot_detected = False

        if spot_detected and print_enabled:
            print("nb_spot_detected: ", nb_spot_detected)
            print("frame_counted: ", frame_counted)
            print("spot_detected: ", spot_detected)
            print(
                f"Track: {self.id}_{self.mother_track_id}_to_{','.join(str(daughter) for daughter in self.daughter_track_ids)}"
            )

        return spot_detected

    def get_bridge_images(
        self, video: np.array, margin: int
    ) -> list[np.array]:
        """
        Generate list of crops around the mid-body.
        First frame is the maximum of cytokinesis frame and mid-body first frame.
        Last frame is the last mid-body frame.

        Parameters
        ----------
        video: np.array
            TYXC
        margin: int
            number of pixels to keep around the mid-body, in all directions

        Returns
        ----------
        bridge_images: list[np.array]
            list of crops around the mid-body, TCYX
        """

        # Case with no mid_body detected
        if not self.mid_body_spots:
            return []

        ordered_mb_frames = sorted(self.mid_body_spots.keys())
        first_mb_frame = ordered_mb_frames[0]
        last_mb_frame = ordered_mb_frames[-1]
        first_frame = max(
            first_mb_frame, self.key_events_frame["cytokinesis"] - 2
        )  # -2?

        bridge_images = []
        for frame in range(first_frame, last_mb_frame + 1):
            min_x = self.position.min_x
            min_y = self.position.min_y

            # Get midbody coordinates
            frame_mid_body = self.mid_body_spots[frame]
            x_pos, y_pos = min_x + frame_mid_body.x, min_y + frame_mid_body.y

            # Extract frame image and crop around the midbody Sir-tubulin
            frame_image = (
                video[frame, :, :, :].squeeze().transpose(2, 0, 1)
            )  # CYX
            crop = smart_cropping(
                frame_image, margin, x_pos, y_pos, pad=True
            )  # CYX
            bridge_images.append(crop)

        return bridge_images

    def adapt_deprecated_attributes(self) -> None:
        """
        Used to adapt deprecated attributes to new ones.
        In particular, x and y instead of position for mid_body_spots.
        """
        # Predicted
        for mid_body_spot in self.mid_body_spots.values():
            if (
                hasattr(mid_body_spot, "position")
                and mid_body_spot.position is not None
            ):
                mid_body_spot.x = mid_body_spot.position[0]
                mid_body_spot.y = mid_body_spot.position[1]
        # Ground truth
        if self.gt_mid_body_spots is None:
            return
        for mid_body_spot in self.gt_mid_body_spots.values():
            if (
                hasattr(mid_body_spot, "position")
                and mid_body_spot.position is not None
            ):
                mid_body_spot.x = mid_body_spot.position[0]
                mid_body_spot.y = mid_body_spot.position[1]

    def get_mid_body_legend(
        self,
    ) -> dict[int, dict[str, Union[int, str]]]:
        """
        Get legend for mid-body spot.
        """
        mid_body_legend = {}
        for frame, mid_body_spot in self.mid_body_spots.items():
            # Get category
            mid_body_category = "undefined"
            for category, category_frame in self.key_events_frame.items():
                if (
                    category_frame > frame
                ):  # make sure current step is before frame
                    continue
                if (
                    mid_body_category
                    in self.key_events_frame  # "undefined" case
                    and self.key_events_frame[mid_body_category]
                    > category_frame
                ):  # make sure we are not going back
                    continue
                mid_body_category = category
            mid_body_legend[frame] = {
                "x": mid_body_spot.x,
                "y": mid_body_spot.y,
                "category": mid_body_category,
            }
        return mid_body_legend



import os
import pickle
from typing import Optional
from matplotlib import pyplot as plt
import numpy as np
import torch
from torch.utils.data import DataLoader
import ast
import csv

import bigfish.stack as stack
from bigfish.plot.utils import save_plot, get_minmax_values
from bigfish.plot.plot_images import _define_patch

from cnn_framework.utils.tools import extract_patterns
from cnn_framework.utils.display_tools import display_progress
from cnn_framework.utils.model_managers.cnn_model_manager import (
    CnnModelManager,
)
from cnn_framework.utils.data_managers.default_data_manager import (
    DefaultDataManager,
)
from cnn_framework.utils.metrics.classification_accuracy import (
    ClassificationAccuracy,
)
from cnn_framework.utils.data_loader_generators.data_loader_generator import (
    collate_dataset_output,
)
from cnn_framework.utils.enum import PredictMode
from cnn_framework.utils.models.resnet_classifier import ResnetClassifier
from cnn_framework.utils.model_params.base_model_params import BaseModelParams

from .cnn_data_set import CnnDataSet
from .hidden_markov_models import HiddenMarkovModel


def re_organize_channels(image: np.ndarray) -> np.ndarray:
    """
    Expect a 4 dimensions image.
    Re-organize channels to get TXYC order.
    """
    if image.ndim != 4:
        raise ValueError("Expect a 4 dimensions image.")

    # Get dimension index of smallest dimension, i.e. channels
    channels_dimension_index = np.argmin(image.shape)
    channels_dimension = image.shape[channels_dimension_index]
    if channels_dimension != 3:
        raise ValueError(
            "Expect 3 channels: SiR-tubulin/MKLP1/Phase contrast, in that order."
        )
    # Put channels at the back
    image = np.moveaxis(image, channels_dimension_index, 3)

    # Get dimension index of second smallest dimension, i.e. time
    second_dimension_index = np.argsort(image.shape)[1]
    # Put time at the front
    image = np.moveaxis(image, second_dimension_index, 0)

    return image


def cell_counter_frame_to_video_frame(
    cell_counter_frame: int, nb_channels=4
) -> int:
    """
    Cell counter index starts at 1, just like Fiji.

    To count frame, it just concatenates all channels.
    For example, with 4 channels, frames 1, 2, 3 and 4 will be frame 1,
    frames 5, 6, 7 and 8 will be frame 2, etc.
    """
    return (cell_counter_frame - 1) // nb_channels


def plot_detection(
    image,
    spots,
    shape="circle",
    radius=3,
    color="red",
    linewidth=1,
    fill=False,
    rescale=False,
    contrast=False,
    title=None,
    framesize=(15, 10),
    remove_frame=True,
    path_output=None,
    ext="png",
    show=True,
):
    """NB: most of this function is copied from the bigfish package.

    Plot detected spots and foci on a 2-d image.

    Parameters
    ----------
    image : np.ndarray
        A 2-d image with shape (y, x).
    spots : list or np.ndarray
        Array with coordinates and shape (nb_spots, 3) or (nb_spots, 2). To
        plot different kind of detected spots with different symbols, use a
        list of arrays.
    shape : list or str, default='circle'
        List of symbols used to localized the detected spots in the image,
        among `circle`, `square` or `polygon`. One symbol per array in `spots`.
        If `shape` is a string, the same symbol is used for every elements of
        'spots'.
    radius : list or int or float, default=3
        List of yx radii of the detected spots, in pixel. One radius per array
        in `spots`. If `radius` is a scalar, the same value is applied for
        every elements of `spots`.
    color : list or str, default='red'
        List of colors of the detected spots. One color per array in `spots`.
        If `color` is a string, the same color is applied for every elements
        of `spots`.
    linewidth : list or int, default=1
        List of widths or width of the border symbol. One integer per array
        in `spots`. If `linewidth` is an integer, the same width is applied
        for every elements of `spots`.
    fill : list or bool, default=False
        List of boolean to fill the symbol of the detected spots. If `fill` is
        a boolean, it is applied for every symbols.
    rescale : bool, default=False
        Rescale pixel values of the image (made by default in matplotlib).
    contrast : bool, default=False
        Contrast image.
    title : str, optional
        Title of the image.
    framesize : tuple, default=(15, 10)
        Size of the frame used to plot with ``plt.figure(figsize=framesize)``.
    remove_frame : bool, default=True
        Remove axes and frame.
    path_output : str, optional
        Path to save the image (without extension).
    ext : str or list, default='png'
        Extension used to save the plot. If it is a list of strings, the plot
        will be saved several times.
    show : bool, default=True
        Show the figure or not.

    """
    # check parameters
    stack.check_array(
        image,
        ndim=2,
        dtype=[
            np.uint8,
            np.uint16,
            np.int32,
            np.int64,
            np.float32,
            np.float64,
        ],
    )
    stack.check_parameter(
        spots=(list),
        shape=(list, str),
        radius=(list, int, float),
        color=(list, str),
        linewidth=(list, int),
        fill=(list, bool),
        rescale=bool,
        contrast=bool,
        title=(str, type(None)),
        framesize=tuple,
        remove_frame=bool,
        path_output=(str, type(None)),
        ext=(str, list),
        show=bool,
    )

    # enlist and format parameters
    n = len(spots)
    if not isinstance(shape, list):
        shape = [shape] * n
    elif isinstance(shape, list) and len(shape) != n:
        raise ValueError(
            "If 'shape' is a list, it should have the same "
            "number of items than spots ({0}).".format(n)
        )
    if not isinstance(radius, list):
        radius = [radius] * n
    elif isinstance(radius, list) and len(radius) != n:
        raise ValueError(
            "If 'radius' is a list, it should have the same "
            "number of items than spots ({0}).".format(n)
        )
    if not isinstance(color, list):
        color = [color] * n
    elif isinstance(color, list) and len(color) != n:
        raise ValueError(
            "If 'color' is a list, it should have the same "
            "number of items than spots ({0}).".format(n)
        )
    if not isinstance(linewidth, list):
        linewidth = [linewidth] * n
    elif isinstance(linewidth, list) and len(linewidth) != n:
        raise ValueError(
            "If 'linewidth' is a list, it should have the same "
            "number of items than spots ({0}).".format(n)
        )
    if not isinstance(fill, list):
        fill = [fill] * n
    elif isinstance(fill, list) and len(fill) != n:
        raise ValueError(
            "If 'fill' is a list, it should have the same "
            "number of items than spots ({0}).".format(n)
        )

    # plot
    fig, ax = plt.subplots(1, 2, sharex="col", figsize=framesize)

    # image
    if not rescale and not contrast:
        vmin, vmax = get_minmax_values(image)
        ax[0].imshow(image, vmin=vmin, vmax=vmax)
    elif rescale and not contrast:
        ax[0].imshow(image)
    else:
        if image.dtype not in [np.int64, bool]:
            image = stack.rescale(image, channel_to_stretch=0)
        ax[0].imshow(image, cmap="gray")

    # spots
    if not rescale and not contrast:
        vmin, vmax = get_minmax_values(image)
        ax[1].imshow(image, vmin=vmin, vmax=vmax)
    elif rescale and not contrast:
        ax[1].imshow(image)
    else:
        if image.dtype not in [np.int64, bool]:
            image = stack.rescale(image, channel_to_stretch=0)
        ax[1].imshow(image, cmap="gray")

    for i, coordinates_2d in enumerate(spots):
        # plot symbols
        patch = _define_patch(
            coordinates_2d[1],
            coordinates_2d[0],
            shape[i],
            radius[i],
            color[i],
            linewidth[i],
            fill[i],
        )
        ax[1].add_patch(patch)

    # titles and frames
    if title is not None:
        ax[0].set_title(title, fontweight="bold", fontsize=10)
        ax[1].set_title("Detection results", fontweight="bold", fontsize=10)
    if remove_frame:
        ax[0].axis("off")
        ax[1].axis("off")
    plt.tight_layout()

    # output
    if path_output is not None:
        save_plot(path_output, ext)
    if show:
        plt.show()
    else:
        plt.close()


def get_annotation_file(video_path, mitosis_track, annotations_files):
    video_file = os.path.basename(video_path).split(".")[0]
    mitosis_id = f"mitosis_{mitosis_track.id}_"
    # Assume that mitosis id is consistent across all Trackmate runs
    # (i.e. possible to use former annotations with new mitoses)
    annotation_path_candidates = extract_patterns(
        annotations_files, ["*" + video_file + "*" + mitosis_id + "*"]
    )
    # Check if video exists
    if len(annotation_path_candidates) == 0:
        return None

    return annotation_path_candidates[0]


def upload_annotations(
    annotations_folder: str,
    video_path: str,
    mitoses_folder: str,
    update_mitoses: Optional[bool] = True,
) -> tuple[int]:
    """
    Function used to upload annotations and perform mid-body detection evaluation.
    """

    # Read video
    video_name = os.path.basename(video_path).split(".")[0]

    mitosis_tracks = []
    # Iterate over "bin" files in exported_mitoses_dir
    for state_path in os.listdir(mitoses_folder):
        # Ignore if not for current video
        if video_name not in state_path:
            continue
        # Load mitosis track
        with open(os.path.join(mitoses_folder, state_path), "rb") as f:
            mitosis_track = pickle.load(f)
            mitosis_track.adapt_deprecated_attributes()

        # Add mitosis track to list
        mitosis_tracks.append(mitosis_track)

    # Get all files in annotations_folder
    annotations_files = []
    for root, _, files in os.walk(annotations_folder):
        for file in files:
            if ".xml" in file:
                annotations_files.append(os.path.join(root, file))

    # Used to evaluate mid-body detection
    mb_detected, mb_not_detected = 0, 0

    # Generate movie for each mitosis and save
    wrong_detections = []
    for i, mitosis_track in enumerate(mitosis_tracks):
        # If annotation is available, load it and update mitosis track
        annotation_file = get_annotation_file(
            video_path, mitosis_track, annotations_files
        )
        if annotation_file is not None:
            mitosis_track.update_mid_body_ground_truth(
                annotation_file, nb_channels=4
            )

            # Save updated mitosis track
            if update_mitoses:
                daughter_track_ids = ",".join(
                    [str(d) for d in mitosis_track.daughter_track_ids]
                )
                state_path = f"{video_name}_mitosis_{mitosis_track.id}_{mitosis_track.mother_track_id}_to_{daughter_track_ids}.bin"
                save_path = os.path.join(
                    mitoses_folder,
                    state_path,
                )
                with open(save_path, "wb") as f:
                    pickle.dump(mitosis_track, f)

            # Evaluate mid-body detection (ignore triple divisions)
            if len(mitosis_track.daughter_track_ids) == 1:
                (
                    is_correctly_detected,
                    percent_detected,
                    average_position_difference,
                ) = mitosis_track.evaluate_mid_body_detection()
                if is_correctly_detected:
                    mb_detected += 1
                else:
                    mb_not_detected += 1
                    wrong_detections.append(
                        {
                            "path": state_path,
                            "percent_detected": percent_detected,
                            "average_position_difference": average_position_difference,
                        }
                    )
                    mitosis_track.evaluate_mid_body_detection()

        display_progress(
            "Update with annotations and evaluate",
            i + 1,
            len(mitosis_tracks),
            additional_message=f"Image {i+1}/{len(mitosis_tracks)}",
        )

    # Print wrong detections
    if len(wrong_detections) > 0:
        print("\nWrong detections:")
        for wrong_detection in wrong_detections:
            print(
                f"{wrong_detection['path']}: detected {wrong_detection['percent_detected']}% with avg distance {wrong_detection['average_position_difference']}"
            )

    if (mb_detected + mb_not_detected) == 0:
        print("No mid-body detection evaluation possible.")
        return 0, 0

    print(
        f"\nMid-body detection evaluation: {mb_detected / (mb_detected + mb_not_detected) * 100:.2f}%"
    )

    return mb_detected, mb_not_detected


def csv_parameters_to_dict(parameters_path: str):
    """
    Read parameters from a csv file and return them as a dictionary.
    Assume to skip some lines if they do not have the right format.
    """
    result_dict = {}
    with open(parameters_path, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        for row in csv_reader:
            splitted_row = row[0].split(";", 1)
            if len(splitted_row) == 1:  # if there is no data in second column
                continue
            if len(row) > 1:  # if there is a comma in second column
                splitted_row[1] += "," + ",".join(row[1:])
            key, value = splitted_row[0], splitted_row[1]
            if (
                len(value) and value[0] == "[" and value[-1] == "]"
            ):  # transform string to list
                value = ast.literal_eval(value)
            result_dict[key] = value
    return result_dict


def perform_cnn_inference(
    model_path: str,
    images: list[np.array],
    cnn_model_params: BaseModelParams,
    cnn_data_set=CnnDataSet,
    cnn_classifier=ResnetClassifier,
    model_manager=CnnModelManager,
):
    """
    Perform CNN inference on a list of images.

    Parameters
    ----------
    model_path : str
        CNN model path
    images :  list[np.array]
        list[CYX]
    cnn_model_params : BaseModelParams
        model parameters
    cnn_data_set : AbstractDataSet
        dataset to read data
    cnn_classifier : ResnetClassifier
        CNN classifier
    model_manager : CnnModelManager
        model manager

    Returns
    -------
    predictions : list[int]
        predicted classes

    """
    # Metaphase model parameters
    model_parameters = cnn_model_params()
    # Modify parameters for inference
    model_parameters.train_ratio = 0
    model_parameters.val_ratio = 0
    model_parameters.test_ratio = 1
    model_parameters.models_folder = model_path

    # Read csv parameters file
    parameters_file = os.path.join(model_path, "parameters.csv")
    training_parameters = csv_parameters_to_dict(parameters_file)

    # Load pretrained model
    model = cnn_classifier(
        nb_classes=int(training_parameters["nb_classes"]),
        nb_input_channels=len(training_parameters["c_indexes"])
        * len(training_parameters["z_indexes"]),
        encoder_name=training_parameters["encoder_name"],
    )

    map_location = None
    if not torch.cuda.is_available():
        map_location = torch.device("cpu")
        print("No GPU found, using CPU.")
    model.load_state_dict(
        torch.load(
            os.path.join(model_path, f"{model_parameters.name}.pt"),
            map_location=map_location,
        )
    )

    # Test (no sampler to keep order)
    dataset_test = cnn_data_set(
        images,
        is_train=False,
        names=[f"{idx}.ext" for idx in range(len(images))],
        data_manager=DefaultDataManager(),
        params=model_parameters,
    )
    test_dl = DataLoader(
        dataset_test,
        batch_size=model_parameters.batch_size,
        collate_fn=collate_dataset_output,
    )

    manager = model_manager(model, model_parameters, ClassificationAccuracy)

    predictions = manager.predict(
        test_dl,
        predict_mode=PredictMode.GetPrediction,
        nb_images_to_save=0,
    )  # careful, this is scores and not probabilities
    predictions = [int(np.argmax(p)) for p in predictions]

    return predictions


def apply_hmm(hmm_parameters, sequence):
    """
    Correct the sequence of classes using HMM.
    """
    # Define observation sequence
    obs_seq = np.asarray(sequence, dtype=np.int32)

    # Define HMM model & run inference
    model = HiddenMarkovModel(
        hmm_parameters["A"], hmm_parameters["B"], hmm_parameters["pi"]
    )
    states_seq, _ = model.viterbi_inference(obs_seq)

    return states_seq



from enum import Enum


# Enum for the different types of templates
class TemplateType(Enum):
    """
    Enum to represent the different embeddings for bridges classification.
    """

    ALL = 1
    ALL_WITHOUT_HARALICK = 2
    NB_PEAKS = 3
    PEAKS_AND_INTENSITY = 4
    HARALICK = 5
    ALL_ABLATION_1 = 6
    ALL_ABLATION_2 = 7
    ALL_ABLATION_3 = 8
    ALL_ABLATION_4 = 9
    ALL_ABLATION_5 = 10
    ALL_ABLATION_6 = 11
    ALL_ABLATION_7 = 12
    ALL_ABLATION_8 = 13
    AVERAGE_CIRCLE = 14



from cut_detector._widget import results_saving
from cut_detector.data.tools import get_data_path
from cut_detector.widget_functions.save_results import perform_results_saving


def test_open_results_saving_widget():
    # Just try to open the widget
    results_saving()


def test_results_saving():
    perform_results_saving(
        get_data_path("mitoses"),
        show=False,
        save_dir=get_data_path("results"),
        verbose=True,
    )


if __name__ == "__main__":
    test_open_results_saving_widget()
    test_results_saving()



# NB: most of the code here is duplicated from original repo

import numpy as np
import operator


def crop_center(img, bounding):
    start = tuple(
        map(lambda a, da: (a - da + 1) // 2, img.shape, bounding)
    )  # 1 is used to match torch method
    end = tuple(map(operator.add, start, bounding))
    slices = tuple(map(slice, start, end))
    return img[slices]


def get_padding(input_shape, output_shape):
    (_, height, width) = input_shape
    (_, desired_height, desired_width) = output_shape

    # Define margins for both sides
    pad_margin_w_left = abs(desired_width - width) // 2
    pad_margin_w_right = abs(desired_width - width) - pad_margin_w_left

    pad_margin_h_top = abs(desired_height - height) // 2
    pad_margin_h_bottom = abs(desired_height - height) - pad_margin_h_top

    return (
        pad_margin_w_left,
        pad_margin_h_top,
        pad_margin_w_right,
        pad_margin_h_bottom,
    )


def resize_padding(image, output_shape, mode, pad_margin_w, pad_margin_h):
    """
    mode: "min" or "zero"
    """
    channels_to_stack = []
    (depth, height, width) = image.shape
    (_, desired_height, desired_width) = output_shape

    # Crop image if it is too big
    if height > desired_height:
        image = crop_center(image, (depth, desired_height, width))
        height = desired_height
    if width > desired_width:
        image = crop_center(image, (depth, height, desired_width))
        width = desired_width

    raw_padding = get_padding(
        image.shape, output_shape
    )  # left, top, right and bottom

    # Define margins for both sides
    if pad_margin_w is None:
        pad_margin_w_left = raw_padding[0]
        pad_margin_w_right = raw_padding[2]
    else:
        pad_margin_w_left, pad_margin_w_right = pad_margin_w

    if pad_margin_h is None:
        pad_margin_h_top = raw_padding[1]
        pad_margin_h_bottom = raw_padding[3]
    else:
        pad_margin_h_top, pad_margin_h_bottom = pad_margin_h

    # Define padding
    padding = (
        (pad_margin_h_top, pad_margin_h_bottom),
        (pad_margin_w_left, pad_margin_w_right),
    )

    # Pad channels with minimal value of current channel
    for channel_image in range(depth):
        if mode == "min":
            constant_values = np.min(image[channel_image, ...])
        elif mode == "zero":
            constant_values = 0
        else:
            raise ValueError("Unknown mode")
        channel_image = np.pad(
            image[channel_image, ...],
            padding,
            mode="constant",
            constant_values=constant_values,
        )
        channels_to_stack.append(channel_image)

    # Stack padded channels and return
    return np.stack(channels_to_stack)


def resize_image(
    image,
    output_shape=None,
    method="zero",
    pad_margin_w=None,
    pad_margin_h=None,
):
    """
    Expect image to be a 3D numpy array CYX.
    Output_shape is a tuple CYX.
    """

    if method == "min" or method == "zero":
        # Compute output_shape
        if output_shape is None:
            if pad_margin_h is None or pad_margin_w is None:
                raise ValueError(
                    "If output_shape is None, pad_margin_h and pad_margin_w must be specified"
                )
            else:
                output_shape = (
                    None,
                    sum(pad_margin_h) + image.shape[1],
                    sum(pad_margin_w) + image.shape[2],
                )
        # Protect against output_shape and margins inconsistency
        if output_shape is not None and pad_margin_h is not None:
            assert sum(pad_margin_h) == output_shape[1] - image.shape[1]
        if output_shape is not None and pad_margin_w is not None:
            assert sum(pad_margin_w) == output_shape[2] - image.shape[2]
        # Pad to required_dimensions
        return resize_padding(
            image, output_shape, method, pad_margin_w, pad_margin_h
        )

    raise ValueError(f"Unknown resize method: {method}")


def smart_cropping(
    microscopy_image,
    margin,
    min_x,
    min_y,
    max_x=None,
    max_y=None,
    fade_margin=False,
    pad=False,
):
    """
    Crop microscopy image with smart margin.
    microscopy_image: ..., YX
    """

    # Get image shape
    height, width = microscopy_image.shape[-2:]

    # Define empty rectangle if single point is given
    if max_x is None:
        max_x = min_x
    if max_y is None:
        max_y = min_y

    # Get smart margin
    clipped_min_y = max(min_y - margin, 0)
    clipped_max_y = min(max_y + margin, height)
    clipped_min_x = max(min_x - margin, 0)
    clipped_max_x = min(max_x + margin, width)

    # Crop image
    clipped_image = np.copy(
        microscopy_image[
            ..., clipped_min_y:clipped_max_y, clipped_min_x:clipped_max_x
        ]
    )

    # Fade margin
    if fade_margin:
        margin_positions = (
            [
                [y, x]
                for x in range(min_x - clipped_min_x)
                for y in range(clipped_max_y - clipped_min_y)
            ]  # left margin
            + [
                [y, x + max_x - clipped_min_x]
                for x in range(clipped_max_x - max_x)
                for y in range(clipped_max_y - clipped_min_y)
            ]  # right margin
            + [
                [y, x]
                for y in range(min_y - clipped_min_y)
                for x in range(min_x - clipped_min_x, max_x - clipped_min_x)
            ]  # top margin
            + [
                [y + max_y - clipped_min_y, x]
                for y in range(clipped_max_y - max_y)
                for x in range(min_x - clipped_min_x, max_x - clipped_min_x)
            ]  # bottom margin
        )
        # Fade margin
        for margin_position in margin_positions:
            clipped_image[..., margin_position[0], margin_position[1]] = (
                clipped_image[..., margin_position[0], margin_position[1]] / 2
            )

    # Pad if necessary
    if pad and clipped_image.shape[-1] * clipped_image.shape[-2] != (
        max_x - min_x + 2 * margin
    ) * (max_y - min_y + 2 * margin):
        # Compute padding
        pad_margin_h = (
            clipped_min_y - (min_y - margin),
            max_y + margin - clipped_max_y,
        )
        pad_margin_w = (
            clipped_min_x - (min_x - margin),
            max_x + margin - clipped_max_x,
        )
        # Pad
        clipped_image = resize_image(
            clipped_image,
            pad_margin_h=pad_margin_h,
            pad_margin_w=pad_margin_w,
            method="zero",
        )

    return clipped_image



from __future__ import annotations

from typing import Optional

from .spot import Spot


class MidBodySpot(Spot):
    """
    Mid-body candidate spot
    """

    def __init__(
        self,
        frame: int,
        x: int,
        y: int,
        intensity: Optional[float] = None,
        sir_intensity: Optional[float] = None,
        area: Optional[float] = None,
        circularity: Optional[float] = None,
    ):
        super().__init__(frame, x, y)
        self.intensity = intensity
        self.sir_intensity = sir_intensity
        self.area = area
        self.circularity = circularity

        self.parent_spot: Optional[MidBodySpot] = None
        self.child_spot: Optional[MidBodySpot] = None

        self.track_id: Optional[int] = None



"""
Taken from https://github.com/TreB1eN/HiddenMarkovModel_Pytorch
"""

import torch


class HiddenMarkovModel(object):
    """
    Hidden Markov self Class
    Parameters:
    -----------

    - N: Number of states.
    - T: numpy.array Transition matrix of size N by N
         stores probability from state i to state j.
    - E: numpy.array Emission matrix of size N by T (number of observations)
         stores the probability of observing  O_j  from state  S_i.
    - T0: numpy.array Initial state probabilities of size N.
    """

    def __init__(self, A, B, pi, epsilon=0.001, max_step=10):
        # Max number of iteration
        self.max_step = max_step
        # convergence criteria
        self.epsilon = epsilon
        # Number of possible states
        self.N = A.shape[0]
        # Number of possible observations
        self.M = B.shape[0]
        # Emission probability
        self.B = torch.tensor(B)
        # Transition matrix
        self.A = torch.tensor(A)
        # Initial state vector
        self.pi = torch.tensor(pi)
        # Used for later display
        self.prob_state_1 = []

    def belief_propagation(self, scores):
        return scores.view(-1, 1) + torch.log(self.A)

    def viterbi_inference(self, x):  # x: observing sequence
        """
        Be careful, all comments below are according to the paper and ignore the log operation
        """
        T = len(x)
        shape = [T, self.N]

        # Init_viterbi_variables
        psi = torch.zeros(shape, dtype=torch.float64)
        delta = torch.zeros_like(psi)
        q_star = torch.zeros([shape[0]], dtype=torch.int64)

        # log probability of emission sequence
        obs_prob_full = torch.log(
            self.B[x]
        )  # b_j(O_t) for all j for all t , shape (T, N)

        # initialize with state starting log-priors
        delta[0] = (
            torch.log(self.pi) + obs_prob_full[0]
        )  # pi_i * b_i(O_1) , shape (T)

        for step, obs_prob in enumerate(
            obs_prob_full[1:]
        ):  # note that first observation is ignored
            belief = self.belief_propagation(
                delta[step, :]
            )  # delta_(step)(i) * a_ij for all j
            # the inferred state by maximizing global function
            psi[step + 1] = torch.argmax(
                belief, 0
            )  # arg_max(delta_(step)(i) * a_ij for all j)
            # and update state and score matrices
            delta[step + 1] = (
                torch.max(belief, 0)[0] + obs_prob
            )  # max(delta_(step)(i) * a_ij for all j) * b_j(O_(step+1)), +1 since first observation is ignored

        # infer most likely last state
        q_star[T - 1] = torch.argmax(delta[T - 1, :], 0)

        for step in range(T - 1, 0, -1):
            # for every timestep retrieve inferred state
            state = q_star[step]
            state_prob = psi[step][state]
            q_star[step - 1] = state_prob

        return q_star, torch.exp(delta)  # turn scores back to probabilities



from __future__ import annotations

from typing import Optional
from shapely.geometry.polygon import Polygon
from shapely import distance

from .spot import Spot


class CellSpot(Spot):
    """
    Useful information for a detected cell.
    """

    def __init__(
        self,
        frame: int,
        x: int,
        y: int,
        id_number: int,
        abs_min_x: int,
        abs_max_x: int,
        abs_min_y: int,
        abs_max_y: int,
        spot_points: list[list[int]],  # (x, y)
    ):
        super().__init__(frame, x, y)

        self.id = id_number
        self.abs_min_x = abs_min_x
        self.abs_max_x = abs_max_x
        self.abs_min_y = abs_min_y
        self.abs_max_y = abs_max_y
        self.spot_points = spot_points

        self.track_id: int = -1

        # Phase predicted by model
        self.predicted_phase: Optional[int] = None
        # Corresponding (closest) metaphase spot in track
        self.corresponding_metaphase_spot = None

    def is_stuck_to(self, other_spot: CellSpot, maximum_stuck_distance: float):
        """
        Distance between two spots hulls.
        """
        return (
            distance(
                Polygon(self.spot_points), Polygon(other_spot.spot_points)
            )
            < maximum_stuck_distance
        )



import os
import pickle
from typing import Optional
import numpy as np

from ..factories.mt_cut_detection_factory import MtCutDetectionFactory
from ..models.tools import get_model_path
from ..utils.mitosis_track import MitosisTrack


def perform_mt_cut_detection(
    raw_video: np.ndarray,
    video_name: str,
    exported_mitoses_dir: str,
    hmm_bridges_parameters_file: Optional[str] = os.path.join(
        get_model_path("hmm"), "hmm_bridges_parameters.npz"
    ),
    bridges_mt_cnn_model_path: Optional[str] = get_model_path(
        "bridges_mt_cnn"
    ),
    update_mitoses: Optional[bool] = True,
):
    """

    Parameters
    ----------
    raw_video: np.ndarray
        TYXC

    """
    mitosis_tracks: list[MitosisTrack] = []
    # Iterate over "bin" files in exported_mitoses_dir
    for state_path in os.listdir(exported_mitoses_dir):
        # Ignore if not for current video
        if video_name not in state_path:
            continue
        # Load mitosis track
        with open(os.path.join(exported_mitoses_dir, state_path), "rb") as f:
            mitosis_track: MitosisTrack = pickle.load(f)
            mitosis_track.adapt_deprecated_attributes()

        # Add mitosis track to list
        mitosis_tracks.append(mitosis_track)

    # Perform cut detection
    mt_cut_detector = MtCutDetectionFactory()
    mt_cut_detector.update_mt_cut_detection(
        mitosis_tracks,
        raw_video,
        hmm_bridges_parameters_file,
        bridges_mt_cnn_model_path,
    )

    for mitosis_track in mitosis_tracks:
        # Save updated mitosis track
        if update_mitoses:
            daughter_track_ids = ",".join(
                [str(d) for d in mitosis_track.daughter_track_ids]
            )
            state_path = f"{video_name}_mitosis_{mitosis_track.id}_{mitosis_track.mother_track_id}_to_{daughter_track_ids}.bin"
            save_path = os.path.join(
                exported_mitoses_dir,
                state_path,
            )
            with open(save_path, "wb") as f:
                pickle.dump(mitosis_track, f)



import os
from skimage import io

from cut_detector._widget import micro_tubules_cut_detection
from cut_detector.data.tools import get_data_path
from cut_detector.widget_functions.mt_cut_detection import (
    perform_mt_cut_detection,
)


def test_open_micro_tubules_cut_detection_widget():
    # Just try to open the widget
    micro_tubules_cut_detection()


def test_micro_tubules_cut_detection():

    # Add video
    video = io.imread(
        os.path.join(get_data_path("videos"), "example_video.tif")  # TYXC
    )

    perform_mt_cut_detection(
        video,
        "example_video",
        get_data_path("mitoses"),
        update_mitoses=False,
    )



import os
from skimage import io

from cut_detector._widget import mid_body_detection
from cut_detector.data.tools import get_data_path
from cut_detector.widget_functions.mid_body_detection import (
    perform_mid_body_detection,
)


def test_open_mid_body_detection_widget():
    # Just try to open the widget
    mid_body_detection()


def test_mid_body_detection_widget():

    # Add video
    video = io.imread(
        os.path.join(get_data_path("videos"), "example_video.tif")
    )  # TYXC

    # Run process
    perform_mid_body_detection(
        video,
        "example_video",
        get_data_path("mitoses"),
        get_data_path("tracks"),
        save_dir=None,
        update_mitoses=False,
    )



from enum import IntEnum


class ImpossibleDetection(IntEnum):
    """
    Enum to represent the different types of impossible detections.
    """

    NORMAL = 0
    NO_MID_BODY_DETECTED = -1
    MORE_THAN_TWO_DAUGHTER_TRACKS = -2
    NEAR_BORDER = -3
    NO_MID_BODY_DETECTED_AFTER_CYTOKINESIS = -4
    MT_CUT_AT_CYTOKINESIS = -5
    NO_CUT_DETECTED = -6
    LIGHT_SPOT = -7



import os
from typing import Optional
from skimage import io
import numpy as np
import napari
import pickle

from cut_detector.data.tools import get_data_path
from cut_detector.utils.mitosis_track import MitosisTrack


def main(
    image_path: Optional[str] = get_data_path("videos"),
    mitoses_path: Optional[str] = get_data_path("mitoses"),
):
    # Create a Napari viewer
    viewer = napari.Viewer()

    # Add video
    video = io.imread(os.path.join(image_path, "example_video.tif"))  # TYXC
    viewer.add_image(video[..., 0].squeeze(), name="micro-tubules")
    viewer.add_image(video[..., 1].squeeze(), name="mid-body")
    viewer.add_image(video[..., 2].squeeze(), name="phase contrast")

    # Create a blue rectangle
    rectangle_width = 50
    rectangle_height = 30
    rectangle_color = [0, 0, 1, 0]  # Blue color in RGB

    # Define the rectangle vertices
    vertices = np.array(
        [
            [0, 0],
            [rectangle_width, 0],
            [rectangle_width, rectangle_height],
            [0, rectangle_height],
        ]
    )

    # Create a rectangle layer # affichage rectangle bleu
    viewer.add_shapes(
        data=vertices,
        shape_type="rectangle",
        edge_color="transparent",
        face_color=rectangle_color,
    )

    # Load mitosis tracks  # masques rajoutés qui suivent les cellules
    mitosis_tracks: list[MitosisTrack] = []
    for state_path in os.listdir(mitoses_path):
        with open(os.path.join(mitoses_path, state_path), "rb") as f:
            mitosis_track = pickle.load(
                f
            )  # sauvegarder une instance de classe et la rechercher après
        mitosis_tracks.append(mitosis_track)

    # Colors list
    colors = np.array(
        [
            [
                np.random.randint(0, 255),
                np.random.randint(0, 255),
                np.random.randint(0, 255),
                0.3,
            ]
            for i in range(100)
        ]
    )

    # Iterate over mitosis_tracks
    mask = np.zeros((50, 600, 600, 4))
    for mitosis_track in mitosis_tracks:
        _, mask_movie = mitosis_track.generate_video_movie(video)
        empty_indexes = np.where(mask_movie == 0)
        cell_indexes = np.where(mask_movie == 1)

        mask_movie = np.stack(
            [mask_movie, mask_movie, mask_movie, mask_movie], axis=-1
        )
        mask_movie[empty_indexes] = [0, 0, 0, 0]
        mask_movie[cell_indexes] = [255, 255, 255, 0.3]

        # Add mask_movie to viewer
        mask[
            mitosis_track.min_frame : mitosis_track.max_frame + 1,
            mitosis_track.position.min_y : mitosis_track.position.max_y,
            mitosis_track.position.min_x : mitosis_track.position.max_x,
            :,
        ] = mask_movie

    viewer.add_image(mask, name="masks", rgb=True)

    # Display the Napari viewer
    napari.run()


if __name__ == "__main__":
    main()
